from betty.ancestry import PersonName, Citation, Source
from betty.asyncio import sync
from betty.tests import TemplateTestCase


class TestNameLabel(TemplateTestCase):
    template_string = '{% import \'macro/person.html.j2\' as personMacros %}{{ personMacros.name_label(name, embedded=embedded if embedded is defined else False) }}'

    @sync
    async def test_with_full_name(self):
        name = PersonName('Jane', 'Dough')
        expected = '<span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Jane</span> <span property="foaf:familyName">Dough</span></span>'
        async with self._render(data={
            'name': name,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_individual_name(self):
        name = PersonName('Jane')
        expected = '<span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Jane</span></span>'
        async with self._render(data={
            'name': name,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_affiliation_name(self):
        name = PersonName(None, 'Dough')
        expected = '<span class="person-label" typeof="foaf:Person">… <span property="foaf:familyName">Dough</span></span>'
        async with self._render(data={
            'name': name,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_embedded(self):
        name = PersonName('Jane', 'Dough')
        source = Source()
        citation = Citation(source)
        name.citations.append(citation)
        expected = '<span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Jane</span> <span property="foaf:familyName">Dough</span></span>'
        async with self._render(data={
            'name': name,
            'embedded': True,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_citation(self):
        name = PersonName('Jane')
        source = Source()
        citation = Citation(source)
        name.citations.append(citation)
        expected = '<span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Jane</span></span><a href="#reference-1" class="citation">[1]</a>'
        async with self._render(data={
            'name': name,
        }) as (actual, _):
            self.assertEqual(expected, actual)



import sys
import unittest
from os import path, makedirs
from tempfile import TemporaryDirectory
from time import sleep

import requests

from betty.config import Configuration
from betty.asyncio import sync
from betty.plugin.nginx import Nginx
from betty.plugin.nginx.serve import DockerizedNginxServer
from betty.site import Site
from betty.tests import TestCase


@unittest.skipIf(sys.platform == 'darwin', 'Mac OS does not natively support Docker.')
class DockerizedNginxServerTest(TestCase):
    @sync
    async def test(self):
        content = 'Hello, and welcome to my site!'
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://example.com')
            configuration.plugins[Nginx] = {}
            makedirs(configuration.www_directory_path)
            with open(path.join(configuration.www_directory_path, 'index.html'), 'w') as f:
                f.write(content)
            async with Site(configuration) as site:
                async with DockerizedNginxServer(site) as server:
                    # Wait for the server to start.
                    sleep(1)
                    response = requests.get(server.public_url)
                    self.assertEquals(200, response.status_code)
                    self.assertEquals(content, response.content.decode('utf-8'))



import calendar
import datetime
import gettext
import operator
import os
from contextlib import suppress
from functools import total_ordering
from typing import Optional, Tuple, Union, List

import babel
from babel import dates, Locale


class Localized:
    locale: Optional[str]

    def __init__(self, locale: Optional[str] = None):
        self.locale = locale


class IncompleteDateError(ValueError):
    pass  # pragma: no cover


@total_ordering
class Date:
    year: Optional[int]
    month: Optional[int]
    day: Optional[int]
    fuzzy: bool

    def __init__(self, year: Optional[int] = None, month: Optional[int] = None, day: Optional[int] = None, fuzzy: bool = False):
        self.year = year
        self.month = month
        self.day = day
        self.fuzzy = fuzzy

    def __repr__(self):
        return '<%s.%s(%s, %s, %s)>' % (self.__class__.__module__, self.__class__.__name__, self.year, self.month, self.day)

    @property
    def comparable(self) -> bool:
        return self.year is not None

    @property
    def complete(self) -> bool:
        return self.year is not None and self.month is not None and self.day is not None

    @property
    def parts(self) -> Tuple[Optional[int], Optional[int], Optional[int]]:
        return self.year, self.month, self.day

    def to_range(self) -> 'DateRange':
        if not self.comparable:
            raise ValueError('Cannot convert non-comparable date %s to a date range.' % self)
        if self.month is None:
            month_start = 1
            month_end = 12
        else:
            month_start = month_end = self.month
        if self.day is None:
            day_start = 1
            day_end = calendar.monthrange(self.year, month_end)[1]
        else:
            day_start = day_end = self.day
        return DateRange(Date(self.year, month_start, day_start), Date(self.year, month_end, day_end))

    def _compare(self, other, comparator):
        if not isinstance(other, Date):
            return NotImplemented
        selfish = self
        if not selfish.comparable or not other.comparable:
            return NotImplemented
        if selfish.complete and other.complete:
            return comparator(selfish.parts, other.parts)
        if not other.complete:
            other = other.to_range()
        if not selfish.complete:
            selfish = selfish.to_range()
        return comparator(selfish, other)

    def __contains__(self, other):
        if isinstance(other, Date):
            return self == other
        if isinstance(other, DateRange):
            return self in other
        raise TypeError('Expected to check a %s, but a %s was given' % (type(Datey), type(other)))

    def __lt__(self, other):
        return self._compare(other, operator.lt)

    def __le__(self, other):
        return self._compare(other, operator.le)

    def __eq__(self, other):
        if not isinstance(other, Date):
            return NotImplemented
        return self.parts == other.parts

    def __ge__(self, other):
        return self._compare(other, operator.ge)

    def __gt__(self, other):
        return self._compare(other, operator.gt)


@total_ordering
class DateRange:
    start: Optional[Date]
    start_is_boundary: bool
    end: Optional[Date]
    end_is_boundary: bool

    def __init__(self, start: Optional[Date] = None, end: Optional[Date] = None, start_is_boundary: bool = False, end_is_boundary: bool = False):
        self.start = start
        self.start_is_boundary = start_is_boundary
        self.end = end
        self.end_is_boundary = end_is_boundary

    def __repr__(self):
        return '%s.%s(%s, %s, start_is_boundary=%s, end_is_boundary=%s)' % (self.__class__.__module__, self.__class__.__name__, repr(self.start), repr(self.end), repr(self.start_is_boundary), repr(self.end_is_boundary))

    @property
    def comparable(self) -> bool:
        return self.start is not None and self.start.comparable or self.end is not None and self.end.comparable

    def __contains__(self, other):
        if not self.comparable:
            return False

        if isinstance(other, Date):
            others = [other]
        elif isinstance(other, DateRange):
            if not other.comparable:
                return False
            others = []
            if other.start is not None and other.start.comparable:
                others.append(other.start)
            if other.end is not None and other.end.comparable:
                others.append(other.end)
        else:
            raise TypeError('Expected to check a %s, but a %s was given' % (type(Datey), type(other)))

        if self.start is not None and self.end is not None:
            if isinstance(other, DateRange) and (other.start is None or other.end is None):
                if other.start is None:
                    return self.start <= other.end or self.end <= other.end
                if other.end is None:
                    return self.start >= other.start or self.end >= other.start
            for another in others:
                if self.start <= another <= self.end:
                    return True
            if isinstance(other, DateRange):
                for selfdate in [self.start, self.end]:
                    if other.start <= selfdate <= other.end:
                        return True

        elif self.start is not None:
            # Two date ranges with start dates only always overlap.
            if isinstance(other, DateRange) and other.end is None:
                return True

            for other in others:
                if self.start <= other:
                    return True
        elif self.end is not None:
            # Two date ranges with end dates only always overlap.
            if isinstance(other, DateRange) and other.start is None:
                return True

            for other in others:
                if other <= self.end:
                    return True

    def __lt__(self, other):
        if not isinstance(other, (Date, DateRange)):
            return NotImplemented

        if not self.comparable or not other.comparable:
            return NotImplemented

        self_has_start = self.start is not None and self.start.comparable
        self_has_end = self.end is not None and self.end.comparable

        if isinstance(other, DateRange):
            other_has_start = other.start is not None and other.start.comparable
            other_has_end = other.end is not None and other.end.comparable

            if self_has_start and other_has_start:
                if self.start == other.start:
                    # If both end dates are missing or incomparable, we consider them equal.
                    if (self.end is None or not self.end.comparable) and (other.end is None or other.end.comparable):
                        return False
                    if self_has_end and other_has_end:
                        return self.end < other.end
                    return other.end is None
                return self.start < other.start

            if self_has_start:
                return self.start < other.end

            if other_has_start:
                return self.end <= other.start

            return self.end < other.end

        if self_has_start:
            return self.start < other
        if self_has_end:
            return self.end <= other

    def __eq__(self, other):
        if isinstance(other, Date):
            return False

        if not isinstance(other, DateRange):
            return NotImplemented
        return (self.start, self.end, self.start_is_boundary, self.end_is_boundary) == (other.start, other.end, other.start_is_boundary, other.end_is_boundary)


Datey = Union[Date, DateRange]


class Translations(gettext.NullTranslations):
    _KEYS = {'_', 'gettext', 'ngettext', 'lgettext', 'lngettext'}

    def __init__(self, fallback: gettext.NullTranslations):
        gettext.NullTranslations.__init__(self)
        self._fallback = fallback
        self._previous_context = {}

    def __enter__(self):
        self.install()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.uninstall()

    def install(self, names=None):
        import builtins
        self._previous_context = {
            key: value for key, value in builtins.__dict__.items() if key in self._KEYS}
        self._fallback.install(self._KEYS)

    def uninstall(self):
        import builtins
        for key in self._KEYS:
            # The function may not have been installed.
            with suppress(KeyError):
                del builtins.__dict__[key]
        builtins.__dict__.update(self._previous_context)


def negotiate_locale(preferred_locale: str, available_locales: List[str]) -> Optional[str]:
    negotiated_locale = babel.negotiate_locale([preferred_locale], available_locales)
    if negotiated_locale is not None:
        return negotiated_locale
    preferred_locale = preferred_locale.split('-', 1)[0]
    for available_locale in available_locales:
        negotiated_locale = babel.negotiate_locale([preferred_locale], [available_locale.split('-', 1)[0]])
        if negotiated_locale is not None:
            return available_locale


def negotiate_localizeds(preferred_locale: str, localizeds: List[Localized]) -> Optional[Localized]:
    negotiated_locale = negotiate_locale(preferred_locale, [localized.locale for localized in localizeds if localized.locale is not None])
    if negotiated_locale is not None:
        for localized in localizeds:
            if localized.locale == negotiated_locale:
                return localized
    for localized in localizeds:
        if localized.locale is None:
            return localized
    with suppress(IndexError):
        return localizeds[0]


def open_translations(locale: str, directory_path: str) -> Optional[gettext.GNUTranslations]:
    try:
        with open(os.path.join(directory_path, 'locale', locale.replace('-', '_'), 'LC_MESSAGES', 'betty.mo'), 'rb') as f:
            return gettext.GNUTranslations(f)
    except FileNotFoundError:
        return None


def format_datey(date: Datey, locale: str) -> str:
    """
    Formats a datey value into a human-readable string.

    This requires gettext translations to be installed. Use the Translations class to do this temporarily.
    """
    try:
        if isinstance(date, Date):
            return format_date(date, locale)
        return format_date_range(date, locale)
    except IncompleteDateError:
        return _('unknown date')


_FORMAT_DATE_FORMATTERS = {
    (True,): lambda: _('around %(date)s'),
    (False,): lambda: _('%(date)s'),
}


def format_date(date: Date, locale: str) -> str:
    return _FORMAT_DATE_FORMATTERS[(date.fuzzy,)]() % {
        'date': _format_date_parts(date, locale),
    }


_FORMAT_DATE_PARTS_FORMATTERS = {
    (True, True, True): lambda: _('MMMM d, y'),
    (True, True, False): lambda: _('MMMM, y'),
    (True, False, False): lambda: _('y'),
    (False, True, True): lambda: _('MMMM d'),
    (False, True, False): lambda: _('MMMM'),
}


def _format_date_parts(date: Date, locale: str) -> str:
    if date is None:
        raise IncompleteDateError('This date is None.')
    try:
        date_parts_format = _FORMAT_DATE_PARTS_FORMATTERS[tuple(map(lambda x: x is not None, date.parts))]()
    except KeyError:
        raise IncompleteDateError('This date does not have enough parts to be rendered.')
    parts = map(lambda x: 1 if x is None else x, date.parts)
    return dates.format_date(datetime.date(*parts), date_parts_format, Locale.parse(locale, '-'))


_FORMAT_DATE_RANGE_FORMATTERS = {
    (False, False, False, False): lambda: _('from %(start_date)s until %(end_date)s'),
    (False, False, False, True): lambda: _('from %(start_date)s until sometime before %(end_date)s'),
    (False, False, True, False): lambda: _('from %(start_date)s until around %(end_date)s'),
    (False, False, True, True): lambda: _('from %(start_date)s until sometime before around %(end_date)s'),
    (False, True, False, False): lambda: _('from sometime after %(start_date)s until %(end_date)s'),
    (False, True, False, True): lambda: _('sometime between %(start_date)s and %(end_date)s'),
    (False, True, True, False): lambda: _('from sometime after %(start_date)s until around %(end_date)s'),
    (False, True, True, True): lambda: _('sometime between %(start_date)s and around %(end_date)s'),
    (True, False, False, False): lambda: _('from around %(start_date)s until %(end_date)s'),
    (True, False, False, True): lambda: _('from around %(start_date)s until sometime before %(end_date)s'),
    (True, False, True, False): lambda: _('from around %(start_date)s until around %(end_date)s'),
    (True, False, True, True): lambda: _('from around %(start_date)s until sometime before around %(end_date)s'),
    (True, True, False, False): lambda: _('from sometime after around %(start_date)s until %(end_date)s'),
    (True, True, False, True): lambda: _('sometime between around %(start_date)s and %(end_date)s'),
    (True, True, True, False): lambda: _('from sometime after around %(start_date)s until around %(end_date)s'),
    (True, True, True, True): lambda: _('sometime between around %(start_date)s and around %(end_date)s'),
    (False, False, None, None): lambda: _('from %(start_date)s'),
    (False, True, None, None): lambda: _('sometime after %(start_date)s'),
    (True, False, None, None): lambda: _('from around %(start_date)s'),
    (True, True, None, None): lambda: _('sometime after around %(start_date)s'),
    (None, None, False, False): lambda: _('until %(end_date)s'),
    (None, None, False, True): lambda: _('sometime before %(end_date)s'),
    (None, None, True, False): lambda: _('until around %(end_date)s'),
    (None, None, True, True): lambda: _('sometime before around %(end_date)s'),
}


def format_date_range(date_range: DateRange, locale: str) -> str:
    formatter_configuration = ()
    formatter_arguments = {}

    try:
        formatter_arguments['start_date'] = _format_date_parts(date_range.start, locale)
        formatter_configuration += (date_range.start.fuzzy, date_range.start_is_boundary)
    except IncompleteDateError:
        formatter_configuration += (None, None)

    try:
        formatter_arguments['end_date'] = _format_date_parts(date_range.end, locale)
        formatter_configuration += (date_range.end.fuzzy, date_range.end_is_boundary)
    except IncompleteDateError:
        formatter_configuration += (None, None)

    if not formatter_arguments:
        raise IncompleteDateError('This date range does not have enough parts to be rendered.')

    return _FORMAT_DATE_RANGE_FORMATTERS[formatter_configuration]() % formatter_arguments



import subprocess as stdsubprocess


def run(args, *remaining_args, **kwargs) -> stdsubprocess.CompletedProcess:
    kwargs['stdout'] = stdsubprocess.PIPE
    kwargs['stderr'] = stdsubprocess.PIPE
    process = stdsubprocess.run(args, *remaining_args, **kwargs)
    if process.returncode != 0:
        raise stdsubprocess.CalledProcessError(process.returncode, args, process.stdout, process.stderr)
    return process



from betty.error import ContextError
from betty.tests import TestCase


class ExternalContextErrorTest(TestCase):
    def test__str__(self):
        message = 'Something went wrong!'
        context = 'Somewhere, at some point...'
        expected = 'Something went wrong!\n- Somewhere, at some point...'
        sut = ContextError(message)
        sut.add_context(context)
        self.assertEquals(expected, str(sut))



from concurrent.futures._base import Executor


class ExceptionRaisingExecutor(Executor):
    def __init__(self, executor: Executor):
        self._executor = executor
        self._futures = []

    def submit(self, *args, **kwargs):
        future = self._executor.submit(*args, **kwargs)
        self._futures.append(future)
        return future

    def map(self, *args, **kwargs):
        return self._executor.map(*args, **kwargs)

    def shutdown(self, *args, **kwargs):
        self._executor.shutdown(*args, **kwargs)
        for future in self._futures:
            future.result()



from typing import Optional, List, Dict

from parameterized import parameterized

from betty.media_type import MediaType, InvalidMediaType
from betty.tests import TestCase


class MediaTypeTest(TestCase):
    @parameterized.expand([
        # The simplest possible media type.
        ('text', 'plain', ['plain'], None, {}, 'text/plain'),
        # A media type with a hyphenated subtype.
        ('multipart', 'form-data', ['form-data'], None, {}, 'multipart/form-data'),
        # A media type with a tree subtype.
        ('application', 'vnd.oasis.opendocument.text', ['vnd', 'oasis', 'opendocument', 'text'], None, {},
         'application/vnd.oasis.opendocument.text'),
        # A media type with a subtype suffix.
        ('application', 'ld+json', ['ld'], 'json', {}, 'application/ld+json'),
        # A media type with a parameter.
        ('text', 'html', ['html'], None, {
            'charset': 'UTF-8',
        }, 'text/html; charset=UTF-8'),
    ])
    def test(self, expected_type: str, expected_subtype: str, expected_subtypes: List[str], expected_suffix: Optional[str], expected_parameters: Dict[str, str], media_type: str):
        sut = MediaType(media_type)
        self.assertEquals(expected_type, sut.type)
        self.assertEquals(expected_subtype, sut.subtype)
        self.assertEquals(expected_subtypes, sut.subtypes)
        self.assertEquals(expected_suffix, sut.suffix)
        self.assertEquals(expected_parameters, sut.parameters)
        self.assertEquals(media_type, str(sut))

    @parameterized.expand([
        ('',),
        ('/',),
        ('text',),
        ('text/',),
    ])
    def test_invalid_type_should_raise_value_error(self, media_type: str):
        with self.assertRaises(InvalidMediaType):
            MediaType(media_type)



import logging
from os import path
from tempfile import TemporaryDirectory

import docker
from docker.errors import DockerException

from betty.plugin.nginx import Nginx, generate_dockerfile_file
from betty.plugin.nginx.docker import Container
from betty.serve import Server, ServerNotStartedError
from betty.site import Site


class DockerizedNginxServer(Server):
    def __init__(self, site: Site):
        self._site = site
        self._container = None
        self._output_directory = None

    async def start(self) -> None:
        logging.getLogger().info('Starting a Dockerized nginx web server...')
        self._output_directory = TemporaryDirectory()
        nginx_configuration_file_path = path.join(self._output_directory.name, 'nginx.conf')
        docker_directory_path = path.join(self._output_directory.name, 'docker')
        dockerfile_file_path = path.join(docker_directory_path, 'Dockerfile')
        async with self._site:
            await self._site.plugins[Nginx].generate_configuration_file(destination_file_path=nginx_configuration_file_path, https=False, www_directory_path='/var/www/betty')
            await generate_dockerfile_file(destination_file_path=dockerfile_file_path)
        self._container = Container(self._site.configuration.www_directory_path, docker_directory_path, nginx_configuration_file_path, 'betty-serve')
        self._container.start()

    async def stop(self) -> None:
        self._container.stop()
        self._output_directory.cleanup()

    @property
    def public_url(self) -> str:
        if self._container is not None:
            return 'http://%s' % self._container.ip
        raise ServerNotStartedError('Cannot determine the public URL if the server has not started yet.')

    @classmethod
    def is_available(cls) -> bool:
        try:
            docker.from_env().info()
            return True
        except DockerException as e:
            logging.getLogger().warning(e)
            return False



import json
from os import path
from tempfile import TemporaryDirectory
from unittest import TestCase

from betty import os, subprocess


class ReadmeTest(TestCase):
    def test_readme_should_contain_cli_help(self):
        with TemporaryDirectory() as betty_site_path:
            configuration = {
                'base_url': 'https://example.com',
                'output': path.join(betty_site_path, 'output'),
            }
            with open(path.join(betty_site_path, 'betty.json'), 'w') as f:
                json.dump(configuration, f)
            with os.ChDir(betty_site_path):
                expected = subprocess.run(['betty', '--help']).stdout
            with open('README.md') as f:
                actual = f.read().encode()
            self.assertIn(expected, actual)



from os import path, makedirs
from tempfile import TemporaryDirectory
from unittest import TestCase

from betty.asyncio import sync
from betty.sass import SassRenderer


class SassRendererTest(TestCase):
    @sync
    async def test_render_file(self) -> None:
        sut = SassRenderer()
        scss = 'outer { inner { display: block; } }'
        expected_css = 'outer inner{display:block}'
        with TemporaryDirectory() as working_directory_path:
            scss_file_path = path.join(working_directory_path, 'betty.scss')
            with open(scss_file_path, 'w') as f:
                f.write(scss)
            await sut.render_file(scss_file_path)
            with open(path.join(working_directory_path, 'betty.css')) as f:
                self.assertEquals(expected_css, f.read().strip())
            self.assertFalse(path.exists(scss_file_path))

    @sync
    async def test_render_file_should_ignore_non_sass_or_scss(self) -> None:
        sut = SassRenderer()
        css = 'outer inner { display: block; }'
        with TemporaryDirectory() as working_directory_path:
            css_file_path = path.join(working_directory_path, 'betty.css')
            with open(css_file_path, 'w') as f:
                f.write(css)
            await sut.render_file(css_file_path)
            with open(path.join(working_directory_path, 'betty.css')) as f:
                self.assertEquals(css, f.read())

    @sync
    async def test_render_tree(self) -> None:
        sut = SassRenderer()
        scss = 'outer { inner { display: block; } }'
        expected_css = 'outer inner{display:block}'
        with TemporaryDirectory() as working_directory_path:
            working_subdirectory_path = path.join(working_directory_path, 'sub')
            makedirs(working_subdirectory_path)
            scss_file_path = path.join(working_subdirectory_path, 'betty.scss')
            with open(scss_file_path, 'w') as f:
                f.write(scss)
            await sut.render_tree(working_directory_path)
            with open(path.join(working_subdirectory_path, 'betty.css')) as f:
                self.assertEquals(expected_css, f.read().strip())
            self.assertFalse(path.exists(scss_file_path))



from os import path
from typing import Optional, Any, Iterable
from urllib.parse import urlparse

from voluptuous import Schema, Required, Maybe

from betty.generate import PostGenerator
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.plugin.nginx.artifact import generate_configuration_file, generate_dockerfile_file
from betty.serve import ServerProvider, Server
from betty.site import Site


class Nginx(Plugin, PostGenerator, ServerProvider):
    configuration_schema: Schema = Schema({
        Required('www_directory_path', default=None): Maybe(str),
        Required('https', default=None): Maybe(bool),
    })

    def __init__(self, site: Site, www_directory_path: Optional[str] = None, https: Optional[bool] = None):
        self._https = https
        self._www_directory_path = www_directory_path
        self._site = site

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(site, configuration['www_directory_path'], configuration['https'])

    @property
    def servers(self) -> Iterable[Server]:
        from betty.plugin.nginx.serve import DockerizedNginxServer

        if DockerizedNginxServer.is_available():
            return [DockerizedNginxServer(self._site)]
        return []

    async def post_generate(self) -> None:
        await self.generate_configuration_file()
        await self._generate_dockerfile_file()

    @property
    def assets_directory_path(self) -> Optional[str]:
        return '%s/assets' % path.dirname(__file__)

    @property
    def https(self) -> bool:
        if self._https is None:
            return self._site.configuration.base_url.startswith('https')
        return self._https

    @property
    def www_directory_path(self) -> str:
        if self._www_directory_path is None:
            return self._site.configuration.www_directory_path
        return self._www_directory_path

    async def generate_configuration_file(self, destination_file_path: Optional[str] = None, **kwargs) -> None:
        kwargs = dict({
            'content_negotiation': self._site.configuration.content_negotiation,
            'https': self._site.plugins[Nginx].https,
            'locale': self._site.locale,
            'locales': self._site.configuration.locales,
            'multilingual': self._site.configuration.multilingual,
            'server_name': urlparse(self._site.configuration.base_url).netloc,
            'www_directory_path': self._site.configuration.www_directory_path,
        }, **kwargs)
        if destination_file_path is None:
            destination_file_path = path.join(self._site.configuration.output_directory_path, 'nginx', 'nginx.conf')
        await generate_configuration_file(destination_file_path, self._site.jinja2_environment, **kwargs)

    async def _generate_dockerfile_file(self) -> None:
        await generate_dockerfile_file(path.join(self._site.configuration.output_directory_path, 'nginx', 'docker', 'Dockerfile'))



from typing import Any

from parameterized import parameterized

from betty.ancestry import Person, Place, File, Source, Identifiable, PlaceName, IdentifiableEvent, \
    IdentifiableSource, IdentifiableCitation, Death
from betty.config import Configuration, LocaleConfiguration
from betty.tests import TestCase
from betty.url import LocalizedPathUrlGenerator, IdentifiableResourceUrlGenerator, SiteUrlGenerator


class LocalizedPathUrlGeneratorTest(TestCase):
    @parameterized.expand([
        ('/', '/'),
        ('/index.html', '/index.html'),
        ('/example', 'example'),
        ('/example', '/example'),
        ('/example/', 'example/'),
        ('/example/', '/example/'),
        ('/example/index.html', 'example/index.html'),
        ('/example/index.html', '/example/index.html'),
    ])
    def test_generate(self, expected: str, resource: str):
        configuration = Configuration('/tmp', 'https://example.com')
        sut = LocalizedPathUrlGenerator(configuration)
        self.assertEquals(expected, sut.generate(resource, 'text/html'))

    @parameterized.expand([
        ('/', 'index.html'),
        ('/', '/index.html'),
        ('/example/', 'example/index.html'),
        ('/example/', '/example/index.html'),
    ])
    def test_generate_with_clean_urls(self, expected: str, resource: str):
        configuration = Configuration('/tmp', 'https://example.com')
        configuration.clean_urls = True
        sut = LocalizedPathUrlGenerator(configuration)
        self.assertEquals(expected, sut.generate(resource, 'text/html'))

    @parameterized.expand([
        ('https://example.com/', '/'),
        ('https://example.com/example', 'example'),
    ])
    def test_generate_absolute(self, expected: str, resource: str):
        configuration = Configuration('/tmp', 'https://example.com')
        sut = LocalizedPathUrlGenerator(configuration)
        self.assertEquals(expected, sut.generate(
            resource, 'text/html', absolute=True))

    def test_generate_with_invalid_value(self):
        configuration = Configuration('/tmp', 'https://example.com')
        sut = LocalizedPathUrlGenerator(configuration)
        with self.assertRaises(ValueError):
            sut.generate(9, 'text/html')

    def test_generate_multilingual(self):
        configuration = Configuration('/tmp', 'https://example.com')
        configuration.locales.clear()
        configuration.locales['nl'] = LocaleConfiguration('nl')
        configuration.locales['en'] = LocaleConfiguration('en')
        sut = LocalizedPathUrlGenerator(configuration)
        self.assertEquals('/nl/index.html',
                          sut.generate('/index.html', 'text/html'))
        self.assertEquals(
            '/en/index.html', sut.generate('/index.html', 'text/html', locale='en'))


class IdentifiableResourceUrlGeneratorTest(TestCase):
    def test_generate(self):
        configuration = Configuration('/tmp', 'https://example.com')
        sut = IdentifiableResourceUrlGenerator(
            configuration, Identifiable, 'prefix/%s/index.%s')
        self.assertEquals('/prefix/I1/index.html',
                          sut.generate(Identifiable('I1'), 'text/html'))

    def test_generate_with_invalid_value(self):
        configuration = Configuration('/tmp', 'https://example.com')
        sut = IdentifiableResourceUrlGenerator(
            configuration, Identifiable, 'prefix/%s/index.html')
        with self.assertRaises(ValueError):
            sut.generate(9, 'text/html')


class SiteUrlGeneratorTest(TestCase):
    @parameterized.expand([
        ('/index.html', '/index.html'),
        ('/person/P1/index.html', Person('P1')),
        ('/event/E1/index.html', IdentifiableEvent('E1', Death())),
        ('/place/P1/index.html', Place('P1', [PlaceName('Place 1')])),
        ('/file/F1/index.html', File('F1', '/tmp')),
        ('/source/S1/index.html', IdentifiableSource('S1', 'Source 1')),
        ('/citation/C1/index.html', IdentifiableCitation('C1', Source('Source 1'))),
    ])
    def test_generate(self, expected: str, resource: Any):
        configuration = Configuration('/tmp', 'https://example.com')
        sut = SiteUrlGenerator(configuration)
        self.assertEquals(expected, sut.generate(resource, 'text/html'))

    def test_generate_with_invalid_value(self):
        configuration = Configuration('/tmp', 'https://example.com')
        sut = SiteUrlGenerator(configuration)
        with self.assertRaises(ValueError):
            sut.generate(9, 'text/html')



import logging
import shutil
import sys
import time
from contextlib import suppress, contextmanager
from functools import wraps
from os import getcwd, path
from typing import Callable, Dict, Optional

import click
from click import get_current_context

import betty
from betty import generate, parse, serve
from betty.config import from_file
from betty.error import UserFacingError
from betty.asyncio import sync
from betty.logging import CliHandler
from betty.site import Site


class CommandValueError(UserFacingError, ValueError):
    pass


class CommandProvider:
    @property
    def commands(self) -> Dict[str, Callable]:
        raise NotImplementedError


@contextmanager
def catch_exceptions():
    try:
        yield
    except Exception as e:
        logger = logging.getLogger()
        if isinstance(e, UserFacingError):
            logger.error(str(e))
        else:
            logger.exception(e)
        sys.exit(1)


def _command(f, is_site_command: bool):
    @wraps(f)
    @catch_exceptions()
    @sync
    async def _command(*args, **kwargs):
        if is_site_command:
            site = get_current_context().obj['site']
            args = (site, *args)
            async with site:
                await f(*args, **kwargs)
        else:
            await f(*args, **kwargs)
    return _command


def global_command(f):
    return _command(f, False)


def site_command(f):
    return _command(f, True)


@catch_exceptions()
@sync
async def _init_ctx(ctx, configuration_file_path: Optional[str] = None) -> None:
    ctx.ensure_object(dict)

    if 'initialized' in ctx.obj:
        return
    ctx.obj['initialized'] = True

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.addHandler(CliHandler())

    ctx.obj['commands'] = {
        'clear-caches': _clear_caches,
    }

    if configuration_file_path is None:
        try_configuration_file_paths = [path.join(getcwd(), 'betty.%s' % extension) for extension in {'json', 'yaml', 'yml'}]
    else:
        try_configuration_file_paths = [configuration_file_path]

    for try_configuration_file_path in try_configuration_file_paths:
        with suppress(FileNotFoundError):
            with open(try_configuration_file_path) as f:
                logger.info('Loading the site from %s.' % try_configuration_file_path)
                configuration = from_file(f)
            site = Site(configuration)
            async with site:
                ctx.obj['commands']['generate'] = _generate
                ctx.obj['commands']['serve'] = _serve
                for plugin in site.plugins.values():
                    if isinstance(plugin, CommandProvider):
                        for command_name, command in plugin.commands.items():
                            ctx.obj['commands'][command_name] = command
            ctx.obj['site'] = site
            return

    if configuration_file_path is not None:
        raise CommandValueError('Configuration file "%s" does not exist.' % configuration_file_path)


class _BettyCommands(click.MultiCommand):
    @catch_exceptions()
    def list_commands(self, ctx):
        _init_ctx(ctx)
        return list(ctx.obj['commands'].keys())

    @catch_exceptions()
    def get_command(self, ctx, cmd_name):
        _init_ctx(ctx)
        with suppress(KeyError):
            return ctx.obj['commands'][cmd_name]


@click.command(cls=_BettyCommands)
@click.option('--configuration', '-c', 'site', is_eager=True, help='The path to a Betty site configuration file. Defaults to betty.json|yaml|yml in the current working directory. This will make additional commands available.', callback=_init_ctx)
def main(site):
    pass


@click.command(help='Clear all caches.')
@global_command
async def _clear_caches():
    with suppress(FileNotFoundError):
        shutil.rmtree(betty._CACHE_DIRECTORY_PATH)
    logging.getLogger().info('All caches cleared.')


@click.command(help='Generate a static site.')
@site_command
async def _generate(site: Site):
    await parse.parse(site)
    await generate.generate(site)


@click.command(help='Serve a generated site.')
@site_command
async def _serve(site: Site):
    if not path.isdir(site.configuration.www_directory_path):
        raise CommandValueError('Web root directory "%s" does not exist.' % site.configuration.www_directory_path)
    async with serve.SiteServer(site):
        while True:
            time.sleep(999999999)



from typing import Set, Type, Any

from betty.ancestry import Ancestry, Person, File, Citation, Source, Event
from betty.functools import walk
from betty.parse import PostParser
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.plugin.privatizer import Privatizer
from betty.site import Site


class AnonymousSource(Source):
    def __init__(self):
        Source.__init__(self)

    @property
    def name(self) -> str:
        return _('Private')

    def replace(self, other: Source) -> None:
        self.citations.append(*other.citations)
        self.contains.append(*other.contains)
        self.files.append(*other.files)


class AnonymousCitation(Citation):
    def __init__(self, source: Source):
        Citation.__init__(self, source)
        self.location = _("A citation is available, but has not been published in order to protect people's privacy")

    def replace(self, other: Citation) -> None:
        self.facts.append(*other.facts)
        self.files.append(*other.files)


def anonymize(ancestry: Ancestry) -> None:
    anonymous_source = AnonymousSource()
    anonymous_citation = AnonymousCitation(anonymous_source)
    for person in ancestry.people.values():
        if person.private:
            anonymize_person(person)
    for event in ancestry.events.values():
        if event.private:
            anonymize_event(event)
    for file in ancestry.files.values():
        if file.private:
            anonymize_file(file)
    for source in ancestry.sources.values():
        if source.private:
            anonymize_source(source, anonymous_source)
    for citation in ancestry.citations.values():
        if citation.private:
            anonymize_citation(citation, anonymous_citation)


def anonymize_person(person: Person) -> None:
    for name in person.names:
        del name.citations
    del person.names
    del person.citations
    del person.files
    for presence in person.presences:
        del presence.event
    del person.presences

    # If a person connects other public people, keep them in the person graph.
    if person.private and not _has_public_descendants(person):
        del person.parents


def _has_public_descendants(person: Person) -> bool:
    for descendant in walk(person, 'children'):
        if not descendant.private:
            return True
    return False


def anonymize_event(event: Event) -> None:
    del event.citations
    del event.files
    del event.presences


def anonymize_file(file: File) -> None:
    del file.resources


def anonymize_source(source: Source, anonymous_source: AnonymousSource) -> None:
    anonymous_source.replace(source)
    del source.citations
    del source.contained_by
    del source.contains
    del source.files


def anonymize_citation(citation: Citation, anonymous_citation: AnonymousCitation) -> None:
    anonymous_citation.replace(citation)
    del citation.facts
    del citation.files
    del citation.source


class Anonymizer(Plugin, PostParser):
    def __init__(self, ancestry: Ancestry):
        self._ancestry = ancestry

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(site.ancestry)

    @classmethod
    def comes_after(cls) -> Set[Type[Plugin]]:
        return {Privatizer}

    async def post_parse(self) -> None:
        anonymize(self._ancestry)



from os import path

from parameterized import parameterized
from voluptuous import Invalid

from betty import os
from betty.tests import TestCase
from betty.voluptuous import Path, Importable


class PathTest(TestCase):
    @parameterized.expand([
        (3,),
        ({},),
        (True,),
    ])
    def test_with_invalid_type_should_raise_invalid(self, path_value):
        with self.assertRaises(Invalid):
            Path()(path_value)

    @parameterized.expand([
        ('/foo/bar', '/foo/bar'),
        (path.join(path.expanduser('~'), 'foo', 'bar'), '~/foo/bar'),
        (path.join(path.expanduser('~'), 'foo', 'bar'), './foo/bar'),
        (path.join(path.expanduser('~'), 'bar'), './foo/../bar'),
    ])
    def test_with_path_should_return(self, expected: str, path_value: str):
        with os.ChDir(path.expanduser('~')):
            self.assertEqual(expected, Path()(path_value))


class ImportableTest(TestCase):
    @parameterized.expand([
        (3,),
        ({},),
        (True,),
    ])
    def test_with_invalid_type_should_raise_invalid(self, importable_value):
        with self.assertRaises(Invalid):
            Importable()(importable_value)

    def test_with_unknown_module_should_raise_invalid(self):
        with self.assertRaises(Invalid):
            Importable()('foo.bar.Baz')

    def test_with_unknown_type_should_raise_invalid(self):
        with self.assertRaises(Invalid):
            Importable()('%s.Foo' % self.__module__)

    def test_with_importable_should_return(self):
        self.assertEqual(self.__class__, Importable()('%s.%s' % (self.__module__, self.__class__.__name__)))



from importlib import import_module
from typing import Any


def import_any(fully_qualified_type_name: str) -> Any:
    try:
        module_name, type_name = fully_qualified_type_name.rsplit('.', 1)
        return getattr(import_module(module_name), type_name)
    except (AttributeError, ValueError):
        raise ImportError('Cannot import "%s".' % fully_qualified_type_name)



import os
from os import mkdir
from os.path import join, dirname
from tempfile import TemporaryDirectory

from betty.fs import iterfiles, FileSystem, hashfile
from betty.asyncio import sync
from betty.tests import TestCase


class IterfilesTest(TestCase):
    @sync
    async def test_iterfiles(self):
        with TemporaryDirectory() as path:
            subdirpath = join(path, 'subdir')
            mkdir(subdirpath)
            open(join(path, 'rootfile'), 'a').close()
            open(join(path, '.hiddenrootfile'), 'a').close()
            open(join(subdirpath, 'subdirfile'), 'a').close()
            actual = [actualpath[len(path) + 1:]
                      async for actualpath in iterfiles(path)]
        expected = [
            '.hiddenrootfile',
            'rootfile',
            'subdir/subdirfile',
        ]
        self.assertCountEqual(expected, actual)


class HashfileTest(TestCase):
    def test_hashfile_with_identical_file(self):
        file_path = join(dirname(dirname(__file__)),
                         'assets', 'public', 'static', 'betty-16x16.png')
        self.assertEquals(hashfile(file_path), hashfile(file_path))

    def test_hashfile_with_different_files(self):
        file_path_1 = join(dirname(dirname(__file__)),
                           'assets', 'public', 'static', 'betty-16x16.png')
        file_path_2 = join(dirname(dirname(__file__)),
                           'assets', 'public', 'static', 'betty-512x512.png')
        self.assertNotEquals(hashfile(file_path_1), hashfile(file_path_2))


class FileSystemTest(TestCase):
    @sync
    async def test_open(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(join(source_path_1, 'apples'), 'w') as f:
                    f.write('apples')
                with open(join(source_path_2, 'apples'), 'w') as f:
                    f.write('notapples')
                with open(join(source_path_1, 'oranges'), 'w') as f:
                    f.write('oranges')
                with open(join(source_path_2, 'bananas'), 'w') as f:
                    f.write('bananas')

                sut = FileSystem(source_path_1, source_path_2)

                with await sut.open('apples') as f:
                    self.assertEquals('apples', f.read())
                with await sut.open('oranges') as f:
                    self.assertEquals('oranges', f.read())
                with await sut.open('bananas') as f:
                    self.assertEquals('bananas', f.read())

                with self.assertRaises(FileNotFoundError):
                    with await sut.open('mangos'):
                        pass

    @sync
    async def test_open_with_first_file_path_alternative_first_source_path(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(join(source_path_1, 'pinkladies'), 'w') as f:
                    f.write('pinkladies')
                with open(join(source_path_2, 'pinkladies'), 'w') as f:
                    f.write('notpinkladies')
                with open(join(source_path_1, 'apples'), 'w') as f:
                    f.write('notpinkladies')
                with open(join(source_path_2, 'apples'), 'w') as f:
                    f.write('notpinkladies')

                sut = FileSystem(source_path_1, source_path_2)

                with await sut.open('pinkladies', 'apples') as f:
                    self.assertEquals('pinkladies', f.read())

    @sync
    async def test_open_with_first_file_path_alternative_second_source_path(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(join(source_path_2, 'pinkladies'), 'w') as f:
                    f.write('pinkladies')
                with open(join(source_path_1, 'apples'), 'w') as f:
                    f.write('notpinkladies')
                with open(join(source_path_2, 'apples'), 'w') as f:
                    f.write('notpinkladies')

                sut = FileSystem(source_path_1, source_path_2)

                with await sut.open('pinkladies', 'apples') as f:
                    self.assertEquals('pinkladies', f.read())

    @sync
    async def test_open_with_second_file_path_alternative_first_source_path(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(join(source_path_1, 'apples'), 'w') as f:
                    f.write('apples')
                with open(join(source_path_2, 'apples'), 'w') as f:
                    f.write('notapples')

                sut = FileSystem(source_path_1, source_path_2)

                with await sut.open('pinkladies', 'apples') as f:
                    self.assertEquals('apples', f.read())

    @sync
    async def test_copy2(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(join(source_path_1, 'apples'), 'w') as f:
                    f.write('apples')
                with open(join(source_path_2, 'apples'), 'w') as f:
                    f.write('notapples')
                with open(join(source_path_1, 'oranges'), 'w') as f:
                    f.write('oranges')
                with open(join(source_path_2, 'bananas'), 'w') as f:
                    f.write('bananas')

                with TemporaryDirectory() as destination_path:
                    sut = FileSystem(source_path_1, source_path_2)

                    await sut.copy2('apples', destination_path)
                    await sut.copy2('oranges', destination_path)
                    await sut.copy2('bananas', destination_path)

                    with await sut.open(join(destination_path, 'apples')) as f:
                        self.assertEquals('apples', f.read())
                    with await sut.open(join(destination_path, 'oranges')) as f:
                        self.assertEquals('oranges', f.read())
                    with await sut.open(join(destination_path, 'bananas')) as f:
                        self.assertEquals('bananas', f.read())

                    with self.assertRaises(FileNotFoundError):
                        await sut.copy2('mangos', destination_path)

    @sync
    async def test_copytree(self):
        with TemporaryDirectory() as source_path_1:
            os.makedirs(join(source_path_1, 'basket'))
            with TemporaryDirectory() as source_path_2:
                os.makedirs(join(source_path_2, 'basket'))
                with open(join(source_path_1, 'basket', 'apples'), 'w') as f:
                    f.write('apples')
                with open(join(source_path_2, 'basket', 'apples'), 'w') as f:
                    f.write('notapples')
                with open(join(source_path_1, 'basket', 'oranges'), 'w') as f:
                    f.write('oranges')
                with open(join(source_path_2, 'basket', 'bananas'), 'w') as f:
                    f.write('bananas')

                with TemporaryDirectory() as destination_path:
                    sut = FileSystem(source_path_1, source_path_2)

                    await sut.copytree('', destination_path)

                    with await sut.open(join(destination_path, 'basket', 'apples')) as f:
                        self.assertEquals('apples', f.read())
                    with await sut.open(join(destination_path, 'basket', 'oranges')) as f:
                        self.assertEquals('oranges', f.read())
                    with await sut.open(join(destination_path, 'basket', 'bananas')) as f:
                        self.assertEquals('bananas', f.read())



from tempfile import TemporaryDirectory
from typing import Optional, Set, Type
from parameterized import parameterized

from betty.ancestry import Person, Presence, Subject, EventType, CreatableDerivableEventType, \
    DerivableEventType, Event, Residence
from betty.config import Configuration
from betty.asyncio import sync
from betty.locale import DateRange, Date, Datey
from betty.parse import parse
from betty.plugin.deriver import derive, Deriver
from betty.site import Site
from betty.tests import TestCase


class Ignored(EventType):
    pass


class ComesBeforeReference(EventType):
    pass


class ComesAfterReference(EventType):
    pass


class ComesBeforeDerivable(DerivableEventType):
    @classmethod
    def comes_before(cls) -> Set[Type['EventType']]:
        return {ComesBeforeReference}


class ComesBeforeCreatableDerivable(CreatableDerivableEventType, ComesBeforeDerivable):
    pass


class ComesAfterDerivable(DerivableEventType):
    @classmethod
    def comes_after(cls) -> Set[Type['EventType']]:
        return {ComesAfterReference}


class ComesAfterCreatableDerivable(CreatableDerivableEventType, ComesAfterDerivable):
    pass


class ComesBeforeAndAfterDerivable(DerivableEventType):
    @classmethod
    def comes_before(cls) -> Set[Type['EventType']]:
        return {Ignored}

    @classmethod
    def comes_after(cls) -> Set[Type['EventType']]:
        return {Ignored}


class ComesBeforeAndAfterCreatableDerivable(CreatableDerivableEventType, DerivableEventType):
    pass


class DeriverTest(TestCase):
    @sync
    async def test_post_parse(self):
        person = Person('P0')
        reference_presence = Presence(person, Subject(), Event(Residence()))
        reference_presence.event.date = Date(1970, 1, 1)

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.plugins[Deriver] = None
            async with Site(configuration) as site:
                site.ancestry.people[person.id] = person
                await parse(site)

        self.assertEquals(3, len(person.presences))
        self.assertEquals(DateRange(None, Date(1970, 1, 1), end_is_boundary=True), person.start.date)
        self.assertEquals(DateRange(Date(1970, 1, 1), start_is_boundary=True), person.end.date)


class DeriveTest(TestCase):
    @parameterized.expand([
        (ComesBeforeDerivable,),
        (ComesBeforeCreatableDerivable,),
        (ComesAfterDerivable,),
        (ComesAfterCreatableDerivable,),
        (ComesBeforeAndAfterDerivable,),
        (ComesBeforeAndAfterCreatableDerivable,),
    ])
    @sync
    async def test_derive_without_events(self, event_type_type: Type[DerivableEventType]):
        person = Person('P0')

        created, updated = derive(person, event_type_type)

        self.assertEquals(0, created)
        self.assertEquals(0, updated)
        self.assertEquals(0, len(person.presences))

    @parameterized.expand([
        (ComesBeforeDerivable,),
        (ComesBeforeCreatableDerivable,),
        (ComesAfterDerivable,),
        (ComesAfterCreatableDerivable,),
        (ComesBeforeAndAfterDerivable,),
        (ComesBeforeAndAfterCreatableDerivable,),
    ])
    @sync
    async def test_derive_create_derivable_events_without_reference_events(self, event_type_type: Type[DerivableEventType]):
        person = Person('P0')
        derivable_event = Event(Ignored())
        Presence(person, Subject(), derivable_event)

        created, updated = derive(person, event_type_type)

        self.assertEquals(0, created)
        self.assertEquals(0, updated)
        self.assertEquals(1, len(person.presences))
        self.assertIsNone(derivable_event.date)

    @parameterized.expand([
        (ComesBeforeDerivable,),
        (ComesBeforeCreatableDerivable,),
        (ComesAfterDerivable,),
        (ComesAfterCreatableDerivable,),
        (ComesBeforeAndAfterDerivable,),
        (ComesBeforeAndAfterCreatableDerivable,),
    ])
    @sync
    async def test_derive_update_derivable_event_without_reference_events(self, event_type_type: Type[DerivableEventType]):
        person = Person('P0')
        Presence(person, Subject(), Event(Ignored()))
        derivable_event = Event(event_type_type())
        Presence(person, Subject(), derivable_event)

        created, updated = derive(person, event_type_type)

        self.assertEquals(0, created)
        self.assertEquals(0, updated)
        self.assertEquals(2, len(person.presences))
        self.assertIsNone(derivable_event.date)

    @parameterized.expand([
        (None, None, None),
        (Date(2000, 1, 1), Date(1970, 1, 1), Date(2000, 1, 1)),
        (Date(1969, 1, 1), Date(1970, 1, 1), Date(1969, 1, 1)),
        (DateRange(Date(2000, 1, 1)), DateRange(Date(1970, 1, 1)), DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1), Date(1970, 1, 1), end_is_boundary=True), DateRange(Date(1970, 1, 1)), DateRange(Date(1969, 1, 1))),
        (DateRange(None, Date(2000, 1, 1)), DateRange(None, Date(1970, 1, 1)), DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), DateRange(None, Date(1970, 1, 1)), DateRange(None, Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), None, DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), None, DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(None, Date(1970, 1, 1), end_is_boundary=True), Date(1970, 1, 1), None),
        (Date(2000, 1, 1), DateRange(Date(1970, 1, 1)), Date(2000, 1, 1)),
        (Date(1969, 1, 1), DateRange(Date(1970, 1, 1)), Date(1969, 1, 1)),
        (DateRange(Date(2000, 1, 1)), DateRange(None, Date(1970, 1, 1)), DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1), Date(1970, 1, 1), end_is_boundary=True), DateRange(None, Date(1970, 1, 1)), DateRange(Date(1969, 1, 1))),
        (DateRange(None, Date(2000, 1, 1)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(None, Date(1969, 1, 1))),
        (DateRange(None, Date(2000, 1, 1)), None, DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), None, DateRange(None, Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), Date(1970, 1, 1), DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), Date(1970, 1, 1), DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(None, Date(1970, 1, 1), end_is_boundary=True), DateRange(Date(1970, 1, 1)), None),
        (Date(2000, 1, 1), DateRange(None, Date(1970, 1, 1)), Date(2000, 1, 1)),
        (Date(1969, 1, 1), DateRange(None, Date(1970, 1, 1)), Date(1969, 1, 1)),
        (DateRange(Date(2000, 1, 1)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1), Date(1970, 1, 1), end_is_boundary=True), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1)), None, DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1)), None, DateRange(Date(1969, 1, 1))),
        (DateRange(None, Date(2000, 1, 1)), Date(1970, 1, 1), DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), Date(1970, 1, 1), DateRange(None, Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), DateRange(Date(1970, 1, 1)), DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), DateRange(Date(1970, 1, 1)), DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(None, Date(1970, 1, 1), end_is_boundary=True), DateRange(None, Date(1970, 1, 1)), None),
        (Date(2000, 1, 1), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), Date(2000, 1, 1)),
        (Date(1969, 1, 1), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), Date(1969, 1, 1)),
        (Date(2000, 1, 1), None, Date(2000, 1, 1)),
        (Date(1969, 1, 1), None, Date(1969, 1, 1)),
        (DateRange(Date(2000, 1, 1)), Date(1970, 1, 1), DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1), Date(1970, 1, 1), end_is_boundary=True), Date(1970, 1, 1), DateRange(Date(1969, 1, 1))),
        (DateRange(None, Date(2000, 1, 1)), DateRange(Date(1970, 1, 1)), DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), DateRange(Date(1970, 1, 1)), DateRange(None, Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), DateRange(None, Date(1970, 1, 1)), DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), DateRange(None, Date(1970, 1, 1)), DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(None, Date(1970, 1, 1), end_is_boundary=True), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), None),
    ])
    @sync
    async def test_derive_update_comes_before_derivable_event(self, expected_datey: Optional[Datey], before_datey: Optional[Datey], derivable_datey: Optional[Datey]):
        expected_updates = 0 if expected_datey == derivable_datey else 1
        person = Person('P0')
        Presence(person, Subject(), Event(Ignored(), Date(0, 0, 0)))
        Presence(person, Subject(), Event(ComesBeforeReference(), before_datey))
        derivable_event = Event(ComesBeforeDerivable(), derivable_datey)
        Presence(person, Subject(), derivable_event)

        created, updated = derive(person, ComesBeforeDerivable)

        self.assertEquals(0, created)
        self.assertEquals(expected_updates, updated)
        self.assertEquals(3, len(person.presences))
        self.assertEquals(expected_datey, derivable_event.date)

    @parameterized.expand([
        (None, None,),
        (DateRange(None, Date(1970, 1, 1), end_is_boundary=True), Date(1970, 1, 1)),
        (None, DateRange(None, None)),
        (DateRange(None, Date(1970, 1, 1), end_is_boundary=True), DateRange(Date(1970, 1, 1))),
        (DateRange(None, Date(1970, 1, 1), end_is_boundary=True), DateRange(None, Date(1970, 1, 1))),
        (DateRange(None, Date(1970, 1, 1), end_is_boundary=True), DateRange(Date(1970, 1, 1), Date(1971, 1, 1))),
    ])
    @sync
    async def test_derive_create_comes_before_derivable_event(self, expected_datey: Optional[Datey], before_datey: Optional[Datey]):
        expected_creations = 0 if expected_datey is None else 1
        person = Person('P0')
        Presence(person, Subject(), Event(Ignored(), Date(0, 0, 0)))
        Presence(person, Subject(), Event(ComesBeforeReference(), before_datey))

        created, updated = derive(person, ComesBeforeCreatableDerivable)

        derived_presences = [presence for presence in person.presences if isinstance(presence.event.type, ComesBeforeCreatableDerivable)]
        self.assertEquals(expected_creations, len(derived_presences))
        if expected_creations:
            derived_presence = derived_presences[0]
            self.assertIsInstance(derived_presence.role, Subject)
            self.assertEquals(expected_datey, derived_presence.event.date)
        self.assertEquals(expected_creations, created)
        self.assertEquals(0, updated)
        self.assertEquals(2 + expected_creations, len(person.presences))

    @parameterized.expand([
        (None, None, None),
        (Date(2000, 1, 1), Date(1970, 1, 1), Date(2000, 1, 1)),
        (Date(1969, 1, 1), Date(1970, 1, 1), Date(1969, 1, 1)),
        (DateRange(Date(2000, 1, 1)), DateRange(Date(1970, 1, 1)), DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1)), DateRange(Date(1970, 1, 1)), DateRange(Date(1969, 1, 1))),
        (DateRange(Date(1970, 1, 1), Date(2000, 1, 1), start_is_boundary=True), DateRange(None, Date(1970, 1, 1)), DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), DateRange(None, Date(1970, 1, 1)), DateRange(None, Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), None, DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), None, DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(Date(1970, 1, 1), start_is_boundary=True), Date(1970, 1, 1), None),
        (Date(2000, 1, 1), DateRange(Date(1970, 1, 1)), Date(2000, 1, 1)),
        (Date(1969, 1, 1), DateRange(Date(1970, 1, 1)), Date(1969, 1, 1)),
        (DateRange(Date(2000, 1, 1)), DateRange(None, Date(1970, 1, 1)), DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1)), DateRange(None, Date(1970, 1, 1)), DateRange(Date(1969, 1, 1))),
        (DateRange(Date(1999, 12, 31), Date(2000, 1, 1), start_is_boundary=True), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(None, Date(1969, 1, 1))),
        (DateRange(None, Date(2000, 1, 1)), None, DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), None, DateRange(None, Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), Date(1970, 1, 1), DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), Date(1970, 1, 1), DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(Date(1970, 1, 1), start_is_boundary=True), DateRange(Date(1970, 1, 1)), None),
        (Date(2000, 1, 1), DateRange(None, Date(1970, 1, 1)), Date(2000, 1, 1)),
        (Date(1969, 1, 1), DateRange(None, Date(1970, 1, 1)), Date(1969, 1, 1)),
        (DateRange(Date(2000, 1, 1)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1)), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), DateRange(Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1)), None, DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1)), None, DateRange(Date(1969, 1, 1))),
        (DateRange(Date(1970, 1, 1), Date(2000, 1, 1), start_is_boundary=True), Date(1970, 1, 1), DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), Date(1970, 1, 1), DateRange(None, Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), DateRange(Date(1970, 1, 1)), DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), DateRange(Date(1970, 1, 1)), DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(Date(1970, 1, 1), start_is_boundary=True), DateRange(None, Date(1970, 1, 1)), None),
        (Date(2000, 1, 1), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), Date(2000, 1, 1)),
        (Date(1969, 1, 1), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), Date(1969, 1, 1)),
        (Date(2000, 1, 1), None, Date(2000, 1, 1)),
        (Date(1969, 1, 1), None, Date(1969, 1, 1)),
        (DateRange(Date(2000, 1, 1)), Date(1970, 1, 1), DateRange(Date(2000, 1, 1))),
        (DateRange(Date(1969, 1, 1)), Date(1970, 1, 1), DateRange(Date(1969, 1, 1))),
        (DateRange(Date(1970, 1, 1), Date(2000, 1, 1), start_is_boundary=True), DateRange(Date(1970, 1, 1)), DateRange(None, Date(2000, 1, 1))),
        (DateRange(None, Date(1969, 1, 1)), DateRange(Date(1970, 1, 1)), DateRange(None, Date(1969, 1, 1))),
        (DateRange(Date(2000, 1, 1), Date(2000, 12, 31)), DateRange(None, Date(1970, 1, 1)), DateRange(Date(2000, 1, 1), Date(2000, 12, 31))),
        (DateRange(Date(1969, 1, 1), Date(1969, 12, 31)), DateRange(None, Date(1970, 1, 1)), DateRange(Date(1969, 1, 1), Date(1969, 12, 31))),
        (DateRange(Date(1999, 12, 31), start_is_boundary=True), DateRange(Date(1970, 1, 1), Date(1999, 12, 31)), None),
    ])
    @sync
    async def test_derive_update_comes_after_derivable_event(self, expected_datey: Optional[Datey], after_datey: Optional[Datey], derivable_datey: Optional[Datey]):
        expected_updates = 0 if expected_datey == derivable_datey else 1
        person = Person('P0')
        Presence(person, Subject(), Event(Ignored(), Date(0, 0, 0)))
        Presence(person, Subject(), Event(ComesAfterReference(), after_datey))
        derivable_event = Event(ComesAfterDerivable(), derivable_datey)
        Presence(person, Subject(), derivable_event)

        created, updated = derive(person, ComesAfterDerivable)

        self.assertEquals(expected_datey, derivable_event.date)
        self.assertEquals(0, created)
        self.assertEquals(expected_updates, updated)
        self.assertEquals(3, len(person.presences))

    @parameterized.expand([
        (None, None),
        (None, Date()),
        (DateRange(Date(1970, 1, 1), start_is_boundary=True), Date(1970, 1, 1)),
        (DateRange(Date(1970, 1, 1), start_is_boundary=True), DateRange(Date(1970, 1, 1))),
        (DateRange(Date(1999, 12, 31), start_is_boundary=True), DateRange(None, Date(1999, 12, 31))),
        (DateRange(Date(1999, 12, 31), start_is_boundary=True), DateRange(Date(1970, 1, 1), Date(1999, 12, 31))),
        (DateRange(Date(1970, 1, 1), start_is_boundary=True), DateRange(Date(1970, 1, 1), Date(1999, 12, 31), end_is_boundary=True)),
    ])
    @sync
    async def test_derive_create_comes_after_derivable_event(self, expected_datey: Optional[Datey], after_datey: Optional[Datey]):
        expected_creations = 0 if expected_datey is None else 1
        person = Person('P0')
        Presence(person, Subject(), Event(Ignored(), Date(0, 0, 0)))
        Presence(person, Subject(), Event(ComesAfterReference(), after_datey))

        created, updated = derive(person, ComesAfterCreatableDerivable)

        derived_presences = [presence for presence in person.presences if isinstance(presence.event.type, ComesAfterCreatableDerivable)]
        self.assertEquals(expected_creations, len(derived_presences))
        if expected_creations:
            derived_presence = derived_presences[0]
            self.assertIsInstance(derived_presence.role, Subject)
            self.assertEquals(expected_datey, derived_presence.event.date)
        self.assertEquals(expected_creations, created)
        self.assertEquals(0, updated)
        self.assertEquals(2 + expected_creations, len(person.presences))



import gzip
import hashlib
import logging
import re
import tarfile
from contextlib import suppress
from os import path
from typing import Tuple, Optional, List, Any

from geopy import Point
from lxml import etree
from lxml.etree import Element
from voluptuous import Schema, IsFile, All

from betty.ancestry import Ancestry, Place, File, Note, PersonName, Presence, PlaceName, Person, Link, HasFiles, \
    HasLinks, HasCitations, IdentifiableEvent, HasPrivacy, IdentifiableSource, IdentifiableCitation, Subject, Witness, \
    Attendee, Birth, Baptism, Adoption, Cremation, Death, Burial, Engagement, Marriage, MarriageAnnouncement, Divorce, \
    DivorceAnnouncement, Residence, Immigration, Emigration, Occupation, Retirement, Correspondence, Confirmation, \
    Funeral, Will, Beneficiary, Enclosure, UnknownEventType, Missing
from betty.config import Path
from betty.fs import makedirs
from betty.locale import DateRange, Datey, Date
from betty.media_type import MediaType
from betty.parse import Parser
from betty.path import rootname
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.site import Site


class _IntermediateAncestry:
    def __init__(self):
        self.notes = {}
        self.files = {}
        self.places = {}
        self.events = {}
        self.people = {}
        self.sources = {}
        self.citations = {}

    def populate(self, ancestry: Ancestry):
        ancestry.files = {
            file.id: file for file in self.files.values()}
        ancestry.people = {
            person.id: person for person in self.people.values()}
        ancestry.places = {place.id: place for place in self.places.values()}
        ancestry.events = {event.id: event for event in self.events.values()}
        ancestry.sources = {
            source.id: source for source in self.sources.values()}
        ancestry.citations = {
            citation.id: citation for citation in self.citations.values()}


class _IntermediatePlace:
    def __init__(self, place: Place, enclosed_by_handles: List[str]):
        self.place = place
        self.enclosed_by_handles = enclosed_by_handles


_NS = {
    'ns': 'http://gramps-project.org/xml/1.7.1/',
}


def _xpath(element, selector: str) -> []:
    return element.xpath(selector, namespaces=_NS)


def _xpath1(element, selector: str) -> Optional:
    elements = element.xpath(selector, namespaces=_NS)
    if elements:
        return elements[0]
    return None


def parse_xml(site: Site, gramps_file_path: str) -> None:
    cache_directory_path = path.join(site.configuration.cache_directory_path, Gramps.name(),
                                     hashlib.md5(gramps_file_path.encode('utf-8')).hexdigest())
    with suppress(FileExistsError):
        makedirs(cache_directory_path)

    logger = logging.getLogger()
    logger.info('Parsing %s...' % gramps_file_path)

    try:
        gramps_file = gzip.open(gramps_file_path)
        try:
            tarfile.open(fileobj=gramps_file).extractall(
                cache_directory_path)
            gramps_file_path = path.join(cache_directory_path, 'data.gramps')
            # Treat the file as a tar archive (*.gpkg) with media and a gzipped XML file (./data.gz/data).
            _parse_tree(site.ancestry, etree.parse(gramps_file_path), cache_directory_path)
        except tarfile.ReadError:
            # Treat the file as a gzipped XML file (*.gramps).
            _parse_tree(site.ancestry, etree.parse(gramps_file), rootname(gramps_file_path))
    except OSError:
        # Treat the file as plain XML (*.gramps).
        _parse_tree(site.ancestry, etree.parse(gramps_file_path), rootname(gramps_file_path))


def _parse_tree(ancestry: Ancestry, tree: etree.ElementTree(), tree_directory_path: str) -> None:
    logger = logging.getLogger()
    database = tree.getroot()
    intermediate_ancestry = _IntermediateAncestry()
    _parse_notes(intermediate_ancestry, database)
    logger.info('Parsed %d notes.' % len(intermediate_ancestry.notes))
    _parse_objects(intermediate_ancestry, database, tree_directory_path)
    logger.info('Parsed %d files.' % len(intermediate_ancestry.files))
    _parse_repositories(intermediate_ancestry, database)
    repository_count = len(intermediate_ancestry.sources)
    logger.info('Parsed %d repositories as sources.' % repository_count)
    _parse_sources(intermediate_ancestry, database)
    logger.info('Parsed %d sources.' % (len(intermediate_ancestry.sources) - repository_count))
    _parse_citations(intermediate_ancestry, database)
    logger.info('Parsed %d citations.' % len(intermediate_ancestry.citations))
    _parse_places(intermediate_ancestry, database)
    logger.info('Parsed %d places.' % len(intermediate_ancestry.places))
    _parse_events(intermediate_ancestry, database)
    logger.info('Parsed %d events.' % len(intermediate_ancestry.events))
    _parse_people(intermediate_ancestry, database)
    logger.info('Parsed %d people.' % len(intermediate_ancestry.people))
    _parse_families(intermediate_ancestry, database)
    intermediate_ancestry.populate(ancestry)


_DATE_PATTERN = re.compile(r'^.{4}((-.{2})?-.{2})?$')
_DATE_PART_PATTERN = re.compile(r'^\d+$')


def _parse_date(element: Element) -> Optional[Datey]:
    dateval_element = _xpath1(element, './ns:dateval[not(@cformat)]')
    if dateval_element is not None:
        dateval_type = _xpath1(dateval_element, './@type')
        if dateval_type is None:
            return _parse_dateval(dateval_element, 'val')
        dateval_type = str(dateval_type)
        if dateval_type == 'about':
            date = _parse_dateval(dateval_element, 'val')
            if date is None:
                return None
            date.fuzzy = True
            return date
        if dateval_type == 'before':
            return DateRange(None, _parse_dateval(dateval_element, 'val'), end_is_boundary=True)
        if dateval_type == 'after':
            return DateRange(_parse_dateval(dateval_element, 'val'), start_is_boundary=True)
    datespan_element = _xpath1(element, './ns:datespan[not(@cformat)]')
    if datespan_element is not None:
        return DateRange(_parse_dateval(datespan_element, 'start'), _parse_dateval(datespan_element, 'stop'))
    daterange_element = _xpath1(element, './ns:daterange[not(@cformat)]')
    if daterange_element is not None:
        return DateRange(_parse_dateval(daterange_element, 'start'), _parse_dateval(daterange_element, 'stop'), start_is_boundary=True, end_is_boundary=True)
    return None


def _parse_dateval(element: Element, value_attribute_name: str) -> Optional[Date]:
    dateval = str(_xpath1(element, './@%s' % value_attribute_name))
    if _DATE_PATTERN.fullmatch(dateval):
        date_parts = [int(part) if _DATE_PART_PATTERN.fullmatch(
            part) and int(part) > 0 else None for part in dateval.split('-')]
        date = Date(*date_parts)
        dateval_quality = _xpath1(element, './@quality')
        if dateval_quality == 'estimated':
            date.fuzzy = True
        return date
    return None


def _parse_notes(ancestry: _IntermediateAncestry, database: Element):
    for element in _xpath(database, './ns:notes/ns:note'):
        _parse_note(ancestry, element)


def _parse_note(ancestry: _IntermediateAncestry, element: Element):
    handle = _xpath1(element, './@handle')
    text = _xpath1(element, './ns:text/text()')
    ancestry.notes[handle] = Note(text)


def _parse_objects(ancestry: _IntermediateAncestry, database: Element, gramps_file_path: str):
    for element in _xpath(database, './ns:objects/ns:object'):
        _parse_object(ancestry, element, gramps_file_path)


def _parse_object(ancestry: _IntermediateAncestry, element: Element, gramps_directory_path):
    handle = _xpath1(element, './@handle')
    entity_id = str(_xpath1(element, './@id'))
    file_element = _xpath1(element, './ns:file')
    file_path = path.join(gramps_directory_path, str(_xpath1(file_element, './@src')))
    file = File(entity_id, file_path)
    file.media_type = MediaType(str(_xpath1(file_element, './@mime')))
    description = str(_xpath1(file_element, './@description'))
    if description:
        file.description = description
    note_handles = _xpath(element, './ns:noteref/@hlink')
    for note_handle in note_handles:
        file.notes.append(ancestry.notes[note_handle])
    _parse_attribute_privacy(file, element, 'attribute')
    ancestry.files[handle] = file


def _parse_people(ancestry: _IntermediateAncestry, database: Element):
    for element in database.xpath('.//*[local-name()="person"]'):
        _parse_person(ancestry, element)


def _parse_person(ancestry: _IntermediateAncestry, element: Element):
    handle = _xpath1(element, './@handle')
    person = Person(str(_xpath1(element, './@id')))

    names = []
    for name_element in _xpath(element, './ns:name'):
        is_alternative = _xpath1(name_element, './@alt') == '1'
        individual_name_element = _xpath1(name_element, './ns:first')
        individual_name = None if individual_name_element is None else individual_name_element.text
        surname_elements = [surname_element for surname_element in _xpath(
            name_element, './ns:surname') if surname_element.text is not None]
        if surname_elements:
            for surname_element in surname_elements:
                if not is_alternative:
                    is_alternative = _xpath1(surname_element, './@prim') == '0'
                affiliation_name = surname_element.text
                surname_prefix = _xpath1(surname_element, './@prefix')
                if surname_prefix is not None:
                    affiliation_name = '%s %s' % (
                        surname_prefix, affiliation_name)
                name = PersonName(individual_name, affiliation_name)
                _parse_citationref(ancestry, name, name_element)
                names.append((name, is_alternative))
        elif individual_name is not None:
            name = PersonName(individual_name)
            _parse_citationref(ancestry, name, name_element)
            names.append((name, is_alternative))
    for name, is_alternative in names:
        if is_alternative:
            person.names.append(name)
        else:
            person.names.prepend(name)

    _parse_eventrefs(ancestry, person, element)
    if str(_xpath1(element, './@priv')) == '1':
        person.private = True

    _parse_citationref(ancestry, person, element)
    _parse_objref(ancestry, person, element)
    _parse_urls(person, element)
    _parse_attribute_privacy(person, element, 'attribute')
    ancestry.people[handle] = person


def _parse_families(ancestry: _IntermediateAncestry, database: Element):
    for element in database.xpath('.//*[local-name()="family"]'):
        _parse_family(ancestry, element)


def _parse_family(ancestry: _IntermediateAncestry, element: Element):
    parents = []

    # Parse the father.
    father_handle = _xpath1(element, './ns:father/@hlink')
    if father_handle:
        father = ancestry.people[father_handle]
        _parse_eventrefs(ancestry, father, element)
        parents.append(father)

    # Parse the mother.
    mother_handle = _xpath1(element, './ns:mother/@hlink')
    if mother_handle:
        mother = ancestry.people[mother_handle]
        _parse_eventrefs(ancestry, mother, element)
        parents.append(mother)

    # Parse the children.
    child_handles = _xpath(element, './ns:childref/@hlink')
    for child_handle in child_handles:
        child = ancestry.people[child_handle]
        for parent in parents:
            parent.children.append(child)


def _parse_eventrefs(ancestry: _IntermediateAncestry, person: Person, element: Element) -> None:
    eventrefs = _xpath(element, './ns:eventref')
    for eventref in eventrefs:
        _parse_eventref(ancestry, person, eventref)


_PRESENCE_ROLE_MAP = {
    'Primary': Subject(),
    'Family': Subject(),
    'Witness': Witness(),
    'Beneficiary': Beneficiary(),
    'Unknown': Attendee(),
}


def _parse_eventref(ancestry: _IntermediateAncestry, person: Person, eventref: Element) -> None:
    event_handle = _xpath1(eventref, './@hlink')
    gramps_presence_role = _xpath1(eventref, './@role')
    role = _PRESENCE_ROLE_MAP[gramps_presence_role] if gramps_presence_role in _PRESENCE_ROLE_MAP else Attendee()
    Presence(person, role, ancestry.events[event_handle])


def _parse_places(ancestry: _IntermediateAncestry, database: Element):
    intermediate_places = {handle: intermediate_place for handle, intermediate_place in
                           [_parse_place(element) for element in database.xpath('.//*[local-name()="placeobj"]')]}
    for intermediate_place in intermediate_places.values():
        for enclosed_by_handle in intermediate_place.enclosed_by_handles:
            Enclosure(intermediate_place.place, intermediate_places[enclosed_by_handle].place)
    ancestry.places = {handle: intermediate_place.place for handle, intermediate_place in
                       intermediate_places.items()}


def _parse_place(element: Element) -> Tuple[str, _IntermediatePlace]:
    handle = _xpath1(element, './@handle')
    names = []
    for name_element in _xpath(element, './ns:pname'):
        # The Gramps language is a single ISO language code, which is a valid BCP 47 locale.
        language = _xpath1(name_element, './@lang')
        date = _parse_date(name_element)
        name = PlaceName(str(_xpath1(name_element, './@value')), locale=language, date=date)
        names.append(name)

    place = Place(_xpath1(element, './@id'), names)

    coordinates = _parse_coordinates(element)
    if coordinates:
        place.coordinates = coordinates

    enclosed_by_handles = _xpath(element, './ns:placeref/@hlink')

    _parse_urls(place, element)

    return handle, _IntermediatePlace(place, enclosed_by_handles)


def _parse_coordinates(element: Element) -> Optional[Point]:
    coord_element = _xpath1(element, './ns:coord')

    if coord_element is None:
        return None

    latitudeval = _xpath1(coord_element, './@lat')
    longitudeval = _xpath1(coord_element, './@long')

    # We could not parse/validate the Gramps coordinates, because they are too freeform.
    with suppress(BaseException):
        return Point(latitudeval, longitudeval)
    return None


def _parse_events(ancestry: _IntermediateAncestry, database: Element):
    for element in database.xpath('.//*[local-name()="event"]'):
        _parse_event(ancestry, element)


_EVENT_TYPE_MAP = {
    'Birth': Birth(),
    'Baptism': Baptism(),
    'Adopted': Adoption(),
    'Cremation': Cremation(),
    'Death': Death(),
    'Funeral': Funeral(),
    'Burial': Burial(),
    'Will': Will(),
    'Engagement': Engagement(),
    'Marriage': Marriage(),
    'Marriage Banns': MarriageAnnouncement(),
    'Divorce': Divorce(),
    'Divorce Filing': DivorceAnnouncement(),
    'Residence': Residence(),
    'Immigration': Immigration(),
    'Emigration': Emigration(),
    'Occupation': Occupation(),
    'Retirement': Retirement(),
    'Correspondence': Correspondence(),
    'Confirmation': Confirmation(),
    'Missing': Missing(),
}


def _parse_event(ancestry: _IntermediateAncestry, element: Element):
    handle = str(_xpath1(element, './@handle'))
    event_id = _xpath1(element, './@id')
    gramps_type = _xpath1(element, './ns:type')

    try:
        event_type = _EVENT_TYPE_MAP[gramps_type.text]
    except KeyError:
        event_type = UnknownEventType()
        logging.getLogger().warning(
            'Betty is unfamiliar with Gramps event "%s"\'s type of "%s". The event was imported, but its type was set to "%s".' % (event_id, gramps_type.text, event_type.label))

    event = IdentifiableEvent(event_id, event_type)

    event.date = _parse_date(element)

    # Parse the event place.
    place_handle = _xpath1(element, './ns:place/@hlink')
    if place_handle:
        event.place = ancestry.places[place_handle]

    # Parse the description.
    description_element = _xpath1(element, './ns:description')
    if description_element is not None:
        event.description = description_element.text

    _parse_objref(ancestry, event, element)
    _parse_citationref(ancestry, event, element)
    _parse_attribute_privacy(event, element, 'attribute')
    ancestry.events[handle] = event


def _parse_repositories(ancestry: _IntermediateAncestry, database: Element) -> None:
    for element in database.xpath('.//*[local-name()="repository"]'):
        _parse_repository(ancestry, element)


def _parse_repository(ancestry: _IntermediateAncestry, element: Element) -> None:
    handle = _xpath1(element, './@handle')

    source = IdentifiableSource(_xpath1(element, './@id'),
                                _xpath1(element, './ns:rname').text)

    _parse_urls(source, element)

    ancestry.sources[handle] = source


def _parse_sources(ancestry: _IntermediateAncestry, database: Element):
    for element in database.xpath('.//*[local-name()="source"]'):
        _parse_source(ancestry, element)


def _parse_source(ancestry: _IntermediateAncestry, element: Element) -> None:
    handle = _xpath1(element, './@handle')

    source = IdentifiableSource(_xpath1(element, './@id'),
                                _xpath1(element, './ns:stitle').text)

    repository_source_handle = _xpath1(element, './ns:reporef/@hlink')
    if repository_source_handle is not None:
        source.contained_by = ancestry.sources[repository_source_handle]

    # Parse the author.
    sauthor_element = _xpath1(element, './ns:sauthor')
    if sauthor_element is not None:
        source.author = sauthor_element.text

    # Parse the publication info.
    spubinfo_element = _xpath1(element, './ns:spubinfo')
    if spubinfo_element is not None:
        source.publisher = spubinfo_element.text

    _parse_objref(ancestry, source, element)
    _parse_attribute_privacy(source, element, 'srcattribute')

    ancestry.sources[handle] = source


def _parse_citations(ancestry: _IntermediateAncestry, database: Element) -> None:
    for element in database.xpath('.//*[local-name()="citation"]'):
        _parse_citation(ancestry, element)


def _parse_citation(ancestry: _IntermediateAncestry, element: Element) -> None:
    handle = _xpath1(element, './@handle')
    source_handle = _xpath1(element, './ns:sourceref/@hlink')

    citation = IdentifiableCitation(_xpath1(element, './@id'),
                                    ancestry.sources[source_handle])

    citation.date = _parse_date(element)
    _parse_objref(ancestry, citation, element)
    _parse_attribute_privacy(citation, element, 'srcattribute')

    page = _xpath1(element, './ns:page')
    if page is not None:
        citation.location = page.text

    ancestry.citations[handle] = citation


def _parse_citationref(ancestry: _IntermediateAncestry, fact: HasCitations, element: Element):
    citation_handles = _xpath(element, './ns:citationref/@hlink')
    for citation_handle in citation_handles:
        fact.citations.append(ancestry.citations[citation_handle])


def _parse_objref(ancestry: _IntermediateAncestry, owner: HasFiles, element: Element):
    file_handles = _xpath(element, './ns:objref/@hlink')
    for file_handle in file_handles:
        owner.files.append(ancestry.files[file_handle])


def _parse_urls(owner: HasLinks, element: Element):
    url_elements = _xpath(element, './ns:url')
    for url_element in url_elements:
        link = Link(str(_xpath1(url_element, './@href')))
        link.relationship = 'external'
        description_element = _xpath1(url_element, './@description')
        if description_element is not None:
            link.label = str(description_element)
        owner.links.add(link)


def _parse_attribute_privacy(resource: HasPrivacy, element: Element, tag: str) -> None:
    privacy_value = _parse_attribute('privacy', element, tag)
    if privacy_value is None:
        return
    if privacy_value == 'private':
        resource.private = True
        return
    if privacy_value == 'public':
        resource.private = False
        return
    logging.getLogger().warning('The betty:privacy Gramps attribute must have a value of "public" or "private", but "%s" was given, which was ignored.' % privacy_value)


def _parse_attribute(name: str, element: Element, tag: str) -> Optional[str]:
    return _xpath1(element, './ns:%s[@type="betty:%s"]/@value' % (tag, name))


class Gramps(Plugin, Parser):
    configuration_schema: Schema = Schema({
        'file': All(str, IsFile(), Path()),
    })

    def __init__(self, site: Site, gramps_file_path: str):
        self._site = site
        self._gramps_file_path = gramps_file_path

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(site, configuration['file'])

    async def parse(self) -> None:
        parse_xml(self._site, self._gramps_file_path)



from typing import Type, Set, Optional, Any

from voluptuous import Schema

from betty.site import Site


NO_CONFIGURATION = None


class Plugin:
    configuration_schema: Schema = Schema(None)

    async def __aenter__(self):
        pass  # pragma: no cover

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass  # pragma: no cover

    @classmethod
    def name(cls) -> str:
        return '%s.%s' % (cls.__module__, cls.__name__)

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        """
        Creates a new instance for a specific site.
        :param site: betty.site.Site
        :param configuration: The configuration must be of the same type as returned by cls.configuration_schema.
        :return: Self
        """
        return cls()

    @classmethod
    def depends_on(cls) -> Set[Type['Plugin']]:
        return set()

    @classmethod
    def comes_after(cls) -> Set[Type['Plugin']]:
        return set()

    @classmethod
    def comes_before(cls) -> Set[Type['Plugin']]:
        return set()

    @property
    def assets_directory_path(self) -> Optional[str]:
        return None



from parameterized import parameterized

from betty.importlib import import_any
from betty.tests import TestCase


class ImportAnyTest(TestCase):
    @parameterized.expand([
        (3,),
        ({},),
        (True,),
    ])
    def test_with_invalid_type_should_raise_invalid(self, importable_value):
        with self.assertRaises(ImportError):
            import_any(importable_value)

    def test_with_unknown_module_should_raise_invalid(self):
        with self.assertRaises(ImportError):
            import_any('foo.bar.Baz')

    def test_with_unknown_type_should_raise_invalid(self):
        with self.assertRaises(ImportError):
            import_any('%s.Foo' % self.__module__)

    def test_with_importable_should_return(self):
        self.assertEqual(self.__class__, import_any('%s.%s' % (self.__module__, self.__class__.__name__)))



import logging
import os
from contextlib import suppress
from json import dump
from os import chmod
from os.path import join
from typing import Iterable, Any

from babel import Locale
from jinja2 import Environment, TemplateNotFound

from betty.fs import makedirs
from betty.json import JSONEncoder
from betty.openapi import build_specification
from betty.site import Site


class PostStaticGenerator:
    async def post_static_generate(self) -> None:
        raise NotImplementedError


class PostGenerator:
    async def post_generate(self) -> None:
        raise NotImplementedError


async def generate(site: Site) -> None:
    logger = logging.getLogger()
    await site.assets.copytree(join('public', 'static'),
                               site.configuration.www_directory_path)
    await site.renderer.render_tree(site.configuration.www_directory_path)
    await site.dispatcher.dispatch(PostStaticGenerator, 'post_static_generate')()
    for locale, locale_configuration in site.configuration.locales.items():
        async with site.with_locale(locale) as site:
            if site.configuration.multilingual:
                www_directory_path = join(
                    site.configuration.www_directory_path, locale_configuration.alias)
            else:
                www_directory_path = site.configuration.www_directory_path

            await site.assets.copytree(join('public', 'localized'), www_directory_path)
            await site.renderer.render_tree(www_directory_path)

            locale_label = Locale.parse(locale, '-').get_display_name()
            await _generate_entity_type(www_directory_path, site.ancestry.files.values(
            ), 'file', site, locale, site.jinja2_environment)
            logger.info('Generated pages for %d files in %s.' %
                        (len(site.ancestry.files), locale_label))
            await _generate_entity_type(www_directory_path, site.ancestry.people.values(
            ), 'person', site, locale, site.jinja2_environment)
            logger.info('Generated pages for %d people in %s.' %
                        (len(site.ancestry.people), locale_label))
            await _generate_entity_type(www_directory_path, site.ancestry.places.values(
            ), 'place', site, locale, site.jinja2_environment)
            logger.info('Generated pages for %d places in %s.' %
                        (len(site.ancestry.places), locale_label))
            await _generate_entity_type(www_directory_path, site.ancestry.events.values(
            ), 'event', site, locale, site.jinja2_environment)
            logger.info('Generated pages for %d events in %s.' %
                        (len(site.ancestry.events), locale_label))
            await _generate_entity_type(www_directory_path, site.ancestry.citations.values(
            ), 'citation', site, locale, site.jinja2_environment)
            logger.info('Generated pages for %d citations in %s.' %
                        (len(site.ancestry.citations), locale_label))
            await _generate_entity_type(www_directory_path, site.ancestry.sources.values(
            ), 'source', site, locale, site.jinja2_environment)
            logger.info('Generated pages for %d sources in %s.' %
                        (len(site.ancestry.sources), locale_label))
            _generate_openapi(www_directory_path, site)
            logger.info('Generated OpenAPI documentation in %s.', locale_label)
    chmod(site.configuration.www_directory_path, 0o755)
    for directory_path, subdirectory_names, file_names in os.walk(site.configuration.www_directory_path):
        for subdirectory_name in subdirectory_names:
            chmod(join(directory_path, subdirectory_name), 0o755)
        for file_name in file_names:
            chmod(join(directory_path, file_name), 0o644)
    await site.dispatcher.dispatch(PostGenerator, 'post_generate')()


def _create_file(path: str) -> object:
    makedirs(os.path.dirname(path))
    return open(path, 'w')


def _create_html_resource(path: str) -> object:
    return _create_file(os.path.join(path, 'index.html'))


def _create_json_resource(path: str) -> object:
    return _create_file(os.path.join(path, 'index.json'))


async def _generate_entity_type(www_directory_path: str, entities: Iterable[Any], entity_type_name: str, site: Site,
                                locale: str, environment: Environment) -> None:
    await _generate_entity_type_list_html(
        www_directory_path, entities, entity_type_name, environment)
    _generate_entity_type_list_json(
        www_directory_path, entities, entity_type_name, site)
    for entity in entities:
        await _generate_entity(www_directory_path, entity,
                               entity_type_name, site, locale, environment)


async def _generate_entity_type_list_html(www_directory_path: str, entities: Iterable[Any], entity_type_name: str,
                                          environment: Environment) -> None:
    entity_type_path = os.path.join(www_directory_path, entity_type_name)
    with suppress(TemplateNotFound):
        template = environment.get_template(
            'page/list-%s.html.j2' % entity_type_name)
        with _create_html_resource(entity_type_path) as f:
            f.write(await template.render_async({
                'page_resource': '/%s/index.html' % entity_type_name,
                'entity_type_name': entity_type_name,
                'entities': entities,
            }))


def _generate_entity_type_list_json(www_directory_path: str, entities: Iterable[Any], entity_type_name: str, site: Site) -> None:
    entity_type_path = os.path.join(www_directory_path, entity_type_name)
    with _create_json_resource(entity_type_path) as f:
        data = {
            '$schema': site.static_url_generator.generate('schema.json#/definitions/%sCollection' % entity_type_name, absolute=True),
            'collection': []
        }
        for entity in entities:
            data['collection'].append(site.localized_url_generator.generate(
                entity, 'application/json', absolute=True))
        dump(data, f)


async def _generate_entity(www_directory_path: str, entity: Any, entity_type_name: str, site: Site, locale: str, environment: Environment) -> None:
    await _generate_entity_html(www_directory_path, entity,
                                entity_type_name, environment)
    _generate_entity_json(www_directory_path, entity,
                          entity_type_name, site, locale)


async def _generate_entity_html(www_directory_path: str, entity: Any, entity_type_name: str, environment: Environment) -> None:
    entity_path = os.path.join(www_directory_path, entity_type_name, entity.id)
    with _create_html_resource(entity_path) as f:
        f.write(await environment.get_template('page/%s.html.j2' % entity_type_name).render_async({
            'page_resource': entity,
            'entity_type_name': entity_type_name,
            entity_type_name: entity,
        }))


def _generate_entity_json(www_directory_path: str, entity: Any, entity_type_name: str, site: Site, locale: str) -> None:
    entity_path = os.path.join(www_directory_path, entity_type_name, entity.id)
    with _create_json_resource(entity_path) as f:
        dump(entity, f, cls=JSONEncoder.get_factory(site, locale))


def _generate_openapi(www_directory_path: str, site: Site) -> None:
    with open(join(www_directory_path, 'api', 'index.json'), 'w') as f:
        dump(build_specification(site), f)



from typing import Any

from parameterized import parameterized

from betty.functools import walk
from betty.tests import TestCase


class WalkTest(TestCase):
    class _Item:
        def __init__(self, child):
            self.child = child

    @parameterized.expand([
        (None,),
        (True,),
        (object(),),
        ([],),
    ])
    def test_with_invalid_attribute(self, item: Any) -> None:
        with self.assertRaises(AttributeError):
            list(walk(item, 'invalid_attribute_name'))

    def test_one_to_one_without_descendants(self) -> None:
        item = self._Item(None)
        actual = walk(item, 'child')
        expected = []
        self.assertEquals(expected, list(actual))

    def test_one_to_one_with_descendants(self) -> None:
        grandchild = self._Item(None)
        child = self._Item(grandchild)
        item = self._Item(child)
        actual = walk(item, 'child')
        expected = [child, grandchild]
        self.assertEquals(expected, list(actual))

    def test_one_to_many_without_descendants(self) -> None:
        item = self._Item([])
        actual = walk(item, 'child')
        expected = []
        self.assertEquals(expected, list(actual))

    def test_with_one_to_many_descendants(self) -> None:
        grandchild = self._Item([])
        child = self._Item([grandchild])
        item = self._Item([child])
        actual = walk(item, 'child')
        expected = [child, grandchild]
        self.assertEquals(expected, list(actual))

    def test_with_mixed_descendants(self) -> None:
        grandchild = self._Item([])
        child = self._Item(grandchild)
        item = self._Item([child])
        actual = walk(item, 'child')
        expected = [child, grandchild]
        self.assertEquals(expected, list(actual))



from betty.ancestry import Event, Place, PlaceName, Citation, Source, Birth
from betty.asyncio import sync
from betty.locale import Date
from betty.tests import TemplateTestCase


class Test(TemplateTestCase):
    template_file = 'event-dimensions.html.j2'

    @sync
    async def test_without_meta(self):
        event = Event(Birth())
        expected = ''
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_date(self):
        event = Event(Birth())
        event.date = Date(1970)
        expected = '1970'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_place(self):
        event = Event(Birth())
        event.place = Place('P0', [PlaceName('The Place')])
        expected = 'in <address><a href="/place/P0/index.html"><span>The Place</span></a></address>'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_place_is_place_context(self):
        event = Event(Birth())
        place = Place('P0', [PlaceName('The Place')])
        event.place = place
        expected = ''
        async with self._render(data={
            'event': event,
            'place_context': place,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_date_and_place(self):
        event = Event(Birth())
        event.date = Date(1970)
        event.place = Place('P0', [PlaceName('The Place')])
        expected = '1970 in <address><a href="/place/P0/index.html"><span>The Place</span></a></address>'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_citation(self):
        event = Event(Birth())
        event.citations.append(Citation(Source('The Source')))
        expected = '<a href="#reference-1" class="citation">[1]</a>'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_embedded(self):
        event = Event(Birth())
        event.date = Date(1970)
        event.place = Place('P0', [PlaceName('The Place')])
        event.citations.append(Citation(Source('The Source')))
        expected = '1970 in <address><span>The Place</span></address>'
        async with self._render(data={
            'event': event,
            'embedded': True,
        }) as (actual, _):
            self.assertEqual(expected, actual)



import os


class ChDir:
    def __init__(self, directory_path: str):
        self._directory_path = directory_path
        self._owd = None

    def __enter__(self) -> None:
        self.change()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.revert()

    def change(self) -> 'ChDir':
        self._owd = os.getcwd()
        os.chdir(self._directory_path)
        return self

    def revert(self) -> None:
        os.chdir(self._owd)



import json
from collections import OrderedDict
from os import path
from typing import Dict, Type, Optional, List, Callable

import yaml
from babel import parse_locale
from voluptuous import Schema, All, Required, Invalid, IsDir, Any, Range

from betty import _CACHE_DIRECTORY_PATH, os
from betty.error import ContextError, UserFacingError
from betty.voluptuous import Path, Importable


class LocaleConfiguration:
    def __init__(self, locale: str, alias: str = None):
        self._locale = locale
        self._alias = alias

    def __repr__(self):
        return '<%s.%s(%s, %s)>' % (self.__class__.__module__, self.__class__.__name__, self.locale, self.alias)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self._locale != other._locale:
            return False
        if self._alias != other._alias:
            return False
        return True

    @property
    def locale(self) -> str:
        return self._locale

    @property
    def alias(self) -> str:
        if self._alias is None:
            return self._locale
        return self._alias


class PluginsConfiguration:
    def __init__(self, plugins_configuration: Dict[Type['Plugin'], Optional[Any]] = None):
        self._plugins_configuration = {}
        if plugins_configuration is not None:
            for plugin_type, plugin_configuration in plugins_configuration.items():
                self[plugin_type] = plugin_configuration

    def __setitem__(self, plugin_type: Type['Plugin'], plugin_configuration: Optional[Any] = None):
        try:
            self._plugins_configuration[plugin_type] = plugin_type.configuration_schema(plugin_configuration)
        except Invalid as e:
            raise ConfigurationValueError(e)

    def __getitem__(self, item):
        return self._plugins_configuration[item]

    def __contains__(self, item):
        return item in self._plugins_configuration

    def __iter__(self):
        yield from self._plugins_configuration.items()

    def __len__(self):
        return len(self._plugins_configuration)


class ThemeConfiguration:
    background_image_id: Optional[str]

    def __init__(self):
        self.background_image_id = None


class Configuration:
    cache_directory_path: str
    content_negotiation: bool
    title: str
    mode: str
    locales: Dict[str, LocaleConfiguration]
    author: Optional[str]
    plugins: PluginsConfiguration
    assets_directory_path: Optional[str]
    theme: ThemeConfiguration
    lifetime_threshold: int

    def __init__(self, output_directory_path: str, base_url: str):
        self.cache_directory_path = _CACHE_DIRECTORY_PATH
        self.output_directory_path = output_directory_path
        self.base_url = base_url.rstrip(
            '/') if not base_url.endswith('://') else base_url
        self._root_path = '/'
        self._clean_urls = False
        self.content_negotiation = False
        self.title = 'Betty'
        self.author = None
        self.plugins = PluginsConfiguration()
        self.mode = 'production'
        self.assets_directory_path = None
        self.locales = OrderedDict()
        default_locale = 'en-US'
        self.locales[default_locale] = LocaleConfiguration(default_locale)
        self.theme = ThemeConfiguration()
        self.lifetime_threshold = 125

    @property
    def www_directory_path(self) -> str:
        return path.join(self.output_directory_path, 'www')

    @property
    def root_path(self) -> str:
        return self._root_path

    @root_path.setter
    def root_path(self, root_path: str):
        if not root_path.endswith('/'):
            root_path += '/'
        self._root_path = root_path

    @property
    def clean_urls(self) -> bool:
        return self._clean_urls or self.content_negotiation

    @clean_urls.setter
    def clean_urls(self, clean_urls: bool):
        self._clean_urls = clean_urls

    @property
    def default_locale(self) -> str:
        return next(iter(self.locales))

    @property
    def multilingual(self) -> bool:
        return len(self.locales) > 1


def _locales_configuration(configuration: List):
    locales_configuration = OrderedDict()
    for locale_configuration in configuration:
        locale = locale_configuration['locale']
        parse_locale(locale, '-')
        locales_configuration[locale] = LocaleConfiguration(
            locale, locale_configuration['alias'] if 'alias' in locale_configuration else None)
    return locales_configuration


def _theme_configuration(config_dict: Dict) -> ThemeConfiguration:
    theme_configuration = ThemeConfiguration()

    for key, value in config_dict.items():
        setattr(theme_configuration, key, value)

    return theme_configuration


def _configuration(config_dict: Dict) -> Configuration:
    configuration = Configuration(
        config_dict.pop('output'), config_dict.pop('base_url'))

    for key, value in config_dict.items():
        setattr(configuration, key, value)

    return configuration


_ConfigurationSchema = Schema(All({
    Required('output'): All(str, Path()),
    'title': str,
    'author': str,
    'locales': All(list, _locales_configuration),
    Required('base_url'): str,
    'root_path': str,
    'clean_urls': bool,
    'content_negotiation': bool,
    'mode': Any('development', 'production'),
    'assets_directory_path': All(str, IsDir(), Path()),
    'plugins': All(dict, lambda x: PluginsConfiguration({Importable()(plugin_type_name): plugin_configuration for plugin_type_name, plugin_configuration in x.items()})),
    Required('theme', default=dict): All({
        'background_image_id': str,
    }, _theme_configuration),
    'lifetime_threshold': All(int, Range(min=1)),
}, _configuration))


class ConfigurationValueError(ContextError, UserFacingError, ValueError):
    pass  # pragma: no cover


def _from_voluptuous(config_builtin: Any) -> Configuration:
    try:
        return _ConfigurationSchema(config_builtin)
    except Invalid as e:
        raise ConfigurationValueError(e)


def _from_json(config_json: str) -> Configuration:
    try:
        return _from_voluptuous(json.loads(config_json))
    except json.JSONDecodeError as e:
        raise ConfigurationValueError('Invalid JSON: %s.' % e)


def _from_yaml(config_yaml: str) -> Configuration:
    try:
        return _from_voluptuous(yaml.safe_load(config_yaml))
    except yaml.YAMLError as e:
        raise ConfigurationValueError('Invalid YAML: %s' % e)


# These factories must take a single argument, which is the configuration in their format, as a string. They must return
# Configuration, or raise ConfigurationValueError.
_from_format_factories: Dict[str, Callable[[str], Configuration]] = {
    '.json': _from_json,
    '.yaml': _from_yaml,
    '.yml': _from_yaml,
}


def from_file(f) -> Configuration:
    file_base_name, file_extension = path.splitext(f.name)
    try:
        factory = _from_format_factories[file_extension]
    except KeyError:
        raise ConfigurationValueError('Unknown file format "%s". Supported formats are: %s.' % (
            file_extension, ', '.join(_from_format_factories.keys())))
    # Change the working directory to allow relative paths to be resolved against the configuration file's directory
    # path.
    with os.ChDir(path.dirname(f.name)):
        try:
            return factory(f.read())
        except ConfigurationValueError as e:
            raise e.add_context('in %s' % path.abspath(f.name))



import gettext
from collections import defaultdict, OrderedDict
from concurrent.futures._base import Executor
from concurrent.futures.process import ProcessPoolExecutor

from jinja2 import Environment

from betty.concurrent import ExceptionRaisingExecutor
from betty.dispatch import Dispatcher
from betty.lock import Locks
from betty.render import Renderer, SequentialRenderer
from betty.sass import SassRenderer

try:
    from contextlib import AsyncExitStack
except ImportError:
    from async_exit_stack import AsyncExitStack
from copy import copy
from os.path import abspath, dirname, join
from typing import Type, Dict

from betty.ancestry import Ancestry
from betty.config import Configuration
from betty.fs import FileSystem
from betty.graph import tsort, Graph
from betty.locale import open_translations, Translations, negotiate_locale
from betty.url import SiteUrlGenerator, StaticPathUrlGenerator, LocalizedUrlGenerator, StaticUrlGenerator


class Site:
    def __init__(self, configuration: Configuration):
        self._site_stack = []
        self._ancestry = Ancestry()
        self._configuration = configuration
        self._assets = FileSystem(
            join(dirname(abspath(__file__)), 'assets'))
        self._dispatcher = Dispatcher()
        self._localized_url_generator = SiteUrlGenerator(configuration)
        self._static_url_generator = StaticPathUrlGenerator(configuration)
        self._locale = None
        self._translations = defaultdict(gettext.NullTranslations)
        self._default_translations = None
        self._plugins = OrderedDict()
        self._plugin_exit_stack = AsyncExitStack()
        self._init_plugins()
        self._init_dispatch_handlers()
        self._init_assets()
        self._init_translations()
        self._jinja2_environment = None
        self._renderer = None
        self._executor = None
        self._locks = Locks()

    async def __aenter__(self):
        if not self._site_stack:
            for plugin in self._plugins.values():
                await self._plugin_exit_stack.enter_async_context(plugin)

        self._default_translations = Translations(self.translations[self.locale])
        self._default_translations.install()

        if self._executor is None:
            self._executor = ExceptionRaisingExecutor(ProcessPoolExecutor())

        self._site_stack.append(self)

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._site_stack.pop()

        self._default_translations.uninstall()

        if not self._site_stack:
            self._executor.shutdown()
            self._executor = None
            await self._plugin_exit_stack.aclose()

    @property
    def locale(self) -> str:
        if self._locale is not None:
            return self._locale
        return self._configuration.default_locale

    def _init_plugins(self) -> None:
        from betty.plugin import NO_CONFIGURATION

        def _extend_plugin_type_graph(graph: Graph, plugin_type: Type['Plugin']):
            dependencies = plugin_type.depends_on()
            # Ensure each plugin type appears in the graph, even if they're isolated.
            graph.setdefault(plugin_type, set())
            for dependency in dependencies:
                seen_dependency = dependency in graph
                graph[dependency].add(plugin_type)
                if not seen_dependency:
                    _extend_plugin_type_graph(graph, dependency)

        plugin_types_graph = defaultdict(set)
        # Add dependencies to the plugin graph.
        for plugin_type, _ in self._configuration.plugins:
            _extend_plugin_type_graph(plugin_types_graph, plugin_type)
        # Now all dependencies have been collected, extend the graph with optional plugin orders.
        for plugin_type, _ in self._configuration.plugins:
            for before in plugin_type.comes_before():
                if before in plugin_types_graph:
                    plugin_types_graph[plugin_type].add(before)
            for after in plugin_type.comes_after():
                if after in plugin_types_graph:
                    plugin_types_graph[after].add(plugin_type)

        for plugin_type in tsort(plugin_types_graph):
            plugin_configuration = self.configuration.plugins[
                plugin_type] if plugin_type in self.configuration.plugins else NO_CONFIGURATION
            plugin = plugin_type.for_site(
                self, plugin_configuration)
            self._plugins[plugin_type] = plugin

    def _init_dispatch_handlers(self) -> None:
        for plugin in self._plugins.values():
            self._dispatcher.append_handler(plugin)

    def _init_assets(self) -> None:
        for plugin in self._plugins.values():
            if plugin.assets_directory_path is not None:
                self._assets.paths.appendleft(
                    plugin.assets_directory_path)
        if self._configuration.assets_directory_path:
            self._assets.paths.appendleft(
                self._configuration.assets_directory_path)

    def _init_translations(self) -> None:
        self._translations['en-US'] = gettext.NullTranslations()
        for locale in self._configuration.locales:
            for assets_path in reversed(self._assets.paths):
                translations = open_translations(locale, assets_path)
                if translations:
                    translations.add_fallback(self._translations[locale])
                    self._translations[locale] = translations

    @property
    def ancestry(self) -> Ancestry:
        return self._ancestry

    @property
    def configuration(self) -> Configuration:
        return self._configuration

    @property
    def plugins(self) -> Dict[Type['Plugin'], 'Plugin']:
        return self._plugins

    @property
    def assets(self) -> FileSystem:
        return self._assets

    @property
    def dispatcher(self) -> Dispatcher:
        return self._dispatcher

    @property
    def localized_url_generator(self) -> LocalizedUrlGenerator:
        return self._localized_url_generator

    @property
    def static_url_generator(self) -> StaticUrlGenerator:
        return self._static_url_generator

    @property
    def translations(self) -> Dict[str, gettext.NullTranslations]:
        return self._translations

    @property
    def jinja2_environment(self) -> Environment:
        if not self._jinja2_environment:
            from betty.jinja2 import BettyEnvironment
            self._jinja2_environment = BettyEnvironment(self)

        return self._jinja2_environment

    @property
    def renderer(self) -> Renderer:
        if not self._renderer:
            from betty.jinja2 import Jinja2Renderer
            self._renderer = SequentialRenderer([
                Jinja2Renderer(self.jinja2_environment, self._configuration),
                SassRenderer(),
            ])

        return self._renderer

    @property
    def executor(self) -> Executor:
        if self._executor is None:
            raise RuntimeError("Cannot get the executor before this site's context is entered.")
        return self._executor

    @property
    def locks(self) -> Locks:
        return self._locks

    def with_locale(self, locale: str) -> 'Site':
        locale = negotiate_locale(locale, list(self.configuration.locales.keys()))
        if locale is None:
            raise ValueError('Locale "%s" is not enabled.' % locale)
        if locale == self.locale:
            return self

        site = copy(self)
        site._locale = locale

        # Clear all locale-dependent lazy-loaded attributes.
        site._jinja2_environment = None
        site._renderer = None

        return site



import io
from logging import Logger, CRITICAL, ERROR, WARNING, INFO, DEBUG
from unittest import TestCase
from unittest.mock import patch

from parameterized import parameterized

from betty.logging import CliHandler


class CliHandlerTest(TestCase):
    @parameterized.expand([
        ('\033[91mSomething went wrong!\033[0m\n',
         'Something went wrong!', CRITICAL),
        ('\033[91mSomething went wrong!\033[0m\n',
         'Something went wrong!', ERROR),
        ('\033[93mSomething went wrong!\033[0m\n',
         'Something went wrong!', WARNING),
        ('\033[92mSomething went wrong!\033[0m\n',
         'Something went wrong!', INFO),
        ('\033[97mSomething went wrong!\033[0m\n',
         'Something went wrong!', DEBUG),
    ])
    @patch('sys.stderr', new_callable=io.StringIO)
    def test_log(self, expected: str, message: str, level: int, stderr: io.StringIO):
        logger = Logger(__name__)
        logger.addHandler(CliHandler())
        logger.log(level, message)
        self.assertEquals(expected, stderr.getvalue())



from concurrent.futures.thread import ThreadPoolExecutor

from betty.concurrent import ExceptionRaisingExecutor
from betty.tests import TestCase


class ExceptionRaisingExecutorTest(TestCase):
    def test_without_exception_should_not_raise(self) -> None:
        def _task():
            return

        with ExceptionRaisingExecutor(ThreadPoolExecutor()) as sut:
            sut.submit(_task)

    def test_with_exception_should_raise(self) -> None:
        def _task():
            raise RuntimeError()

        with self.assertRaises(RuntimeError):
            with ExceptionRaisingExecutor(ThreadPoolExecutor()) as sut:
                sut.submit(_task)



import re
from os.path import join
from tempfile import TemporaryDirectory
from typing import Optional

from betty.config import Configuration, LocaleConfiguration
from betty.asyncio import sync
from betty.generate import generate
from betty.plugin.nginx import Nginx
from betty.site import Site
from betty.tests import TestCase


class NginxTest(TestCase):
    _LEADING_WHITESPACE_PATTERN = re.compile(r'^\s*(.*?)$')

    def _normalize_configuration(self, configuration: str) -> str:
        return '\n'.join(filter(None, map(self._normalize_configuration_line, configuration.splitlines())))

    def _normalize_configuration_line(self, line: str) -> Optional[str]:
        match = self._LEADING_WHITESPACE_PATTERN.fullmatch(line)
        if match is None:
            return None
        return match.group(1)

    async def _assert_configuration_equals(self, expected: str, configuration: Configuration):
        async with Site(configuration) as site:
            await generate(site)
        with open(join(configuration.output_directory_path, 'nginx', 'nginx.conf')) as f:
            actual = f.read()
        self.assertEqual(self._normalize_configuration(expected), self._normalize_configuration(actual))

    @sync
    async def test_post_render_config(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'http://example.com')
            configuration.plugins[Nginx] = {}
            expected = r'''
server {
    listen 80;
    server_name example.com;
    root %s;
    add_header Cache-Control "max-age=86400";
    gzip on;
    gzip_disable "msie6";
    gzip_vary on;
    gzip_types text/css application/javascript application/json application/xml;

    set $media_type_extension html;
    index index.$media_type_extension;

    location / {
        # Handle HTTP error responses.
        error_page 401 /.error/401.$media_type_extension;
        error_page 403 /.error/403.$media_type_extension;
        error_page 404 /.error/404.$media_type_extension;
        location /.error {
            internal;
        }

        try_files $uri $uri/ =404;
    }
}
''' % configuration.www_directory_path
            await self._assert_configuration_equals(expected, configuration)

    @sync
    async def test_post_render_config_with_clean_urls(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'http://example.com')
            configuration.plugins[Nginx] = {}
            configuration.clean_urls = True
            expected = r'''
server {
    listen 80;
    server_name example.com;
    root %s;
    add_header Cache-Control "max-age=86400";
    gzip on;
    gzip_disable "msie6";
    gzip_vary on;
    gzip_types text/css application/javascript application/json application/xml;

    set $media_type_extension html;
    index index.$media_type_extension;

    location / {
        # Handle HTTP error responses.
        error_page 401 /.error/401.$media_type_extension;
        error_page 403 /.error/403.$media_type_extension;
        error_page 404 /.error/404.$media_type_extension;
        location /.error {
            internal;
        }

        try_files $uri $uri/ =404;
    }
}
''' % configuration.www_directory_path
            await self._assert_configuration_equals(expected, configuration)

    @sync
    async def test_post_render_config_multilingual(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'http://example.com')
            configuration.plugins[Nginx] = {}
            configuration.locales.clear()
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            expected = r'''
server {
    listen 80;
    server_name example.com;
    root %s;
    add_header Cache-Control "max-age=86400";
    gzip on;
    gzip_disable "msie6";
    gzip_vary on;
    gzip_types text/css application/javascript application/json application/xml;

    set $media_type_extension html;
    index index.$media_type_extension;

    location ~ ^/(en|nl)(/|$) {
        set $locale $1;

        add_header Content-Language "$locale" always;

        # Handle HTTP error responses.
        error_page 401 /$locale/.error/401.$media_type_extension;
        error_page 403 /$locale/.error/403.$media_type_extension;
        error_page 404 /$locale/.error/404.$media_type_extension;
        location ~ ^/$locale/\.error {
            internal;
        }

        try_files $uri $uri/ =404;
    }
    location @localized_redirect {
        set $locale_alias en;
        return 301 /$locale_alias$uri;
    }
    location / {
        try_files $uri @localized_redirect;
    }
}
''' % configuration.www_directory_path
            await self._assert_configuration_equals(expected, configuration)

    @sync
    async def test_post_render_config_multilingual_with_content_negotiation(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'http://example.com')
            configuration.content_negotiation = True
            configuration.plugins[Nginx] = {}
            configuration.locales.clear()
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            expected = r'''
server {
    listen 80;
    server_name example.com;
    root %s;
    add_header Cache-Control "max-age=86400";
    gzip on;
    gzip_disable "msie6";
    gzip_vary on;
    gzip_types text/css application/javascript application/json application/xml;

    set_by_lua_block $media_type_extension {
        local available_media_types = {'text/html', 'application/json'}
        local media_type_extensions = {}
        media_type_extensions['text/html'] = 'html'
        media_type_extensions['application/json'] = 'json'
        local media_type = require('cone').negotiate(ngx.req.get_headers()['Accept'], available_media_types)
        return media_type_extensions[media_type]
    }
    index index.$media_type_extension;

    location ~ ^/(en|nl)(/|$) {
        set $locale $1;

        add_header Content-Language "$locale" always;

        # Handle HTTP error responses.
        error_page 401 /$locale/.error/401.$media_type_extension;
        error_page 403 /$locale/.error/403.$media_type_extension;
        error_page 404 /$locale/.error/404.$media_type_extension;
        location ~ ^/$locale/\.error {
            internal;
        }

        try_files $uri $uri/ =404;
    }
    location @localized_redirect {
        set_by_lua_block $locale_alias {
            local available_locales = {'en-US', 'nl-NL'}
            local locale_aliases = {}
            locale_aliases['en-US'] = 'en'
            locale_aliases['nl-NL'] = 'nl'
            local locale = require('cone').negotiate(ngx.req.get_headers()['Accept-Language'], available_locales)
            return locale_aliases[locale]
        }
        return 301 /$locale_alias$uri;
    }
    location / {
        try_files $uri @localized_redirect;
    }
}
''' % configuration.www_directory_path
            await self._assert_configuration_equals(expected, configuration)

    @sync
    async def test_post_render_config_with_content_negotiation(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'http://example.com')
            configuration.content_negotiation = True
            configuration.plugins[Nginx] = {}
            expected = r'''
server {
    listen 80;
    server_name example.com;
    root %s;
    add_header Cache-Control "max-age=86400";
    gzip on;
    gzip_disable "msie6";
    gzip_vary on;
    gzip_types text/css application/javascript application/json application/xml;

    set_by_lua_block $media_type_extension {
        local available_media_types = {'text/html', 'application/json'}
        local media_type_extensions = {}
        media_type_extensions['text/html'] = 'html'
        media_type_extensions['application/json'] = 'json'
        local media_type = require('cone').negotiate(ngx.req.get_headers()['Accept'], available_media_types)
        return media_type_extensions[media_type]
    }
    index index.$media_type_extension;

    location / {
        # Handle HTTP error responses.
        error_page 401 /.error/401.$media_type_extension;
        error_page 403 /.error/403.$media_type_extension;
        error_page 404 /.error/404.$media_type_extension;
        location /.error {
            internal;
        }

        try_files $uri $uri/ =404;
    }
}''' % configuration.www_directory_path
            await self._assert_configuration_equals(expected, configuration)

    @sync
    async def test_post_render_config_with_https(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.plugins[Nginx] = {}
            expected = r'''
server {
    listen 80;
    server_name example.com;
    return 301 https://$host$request_uri;
}
server {
    listen 443 ssl http2;
    server_name example.com;
    root %s;
    add_header Strict-Transport-Security "max-age=31536000; includeSubDomains" always;
    add_header Cache-Control "max-age=86400";
    gzip on;
    gzip_disable "msie6";
    gzip_vary on;
    gzip_types text/css application/javascript application/json application/xml;

    set $media_type_extension html;
    index index.$media_type_extension;

    location / {
        # Handle HTTP error responses.
        error_page 401 /.error/401.$media_type_extension;
        error_page 403 /.error/403.$media_type_extension;
        error_page 404 /.error/404.$media_type_extension;
        location /.error {
            internal;
    }

    try_files $uri $uri/ =404;
    }
}
''' % configuration.www_directory_path
            await self._assert_configuration_equals(expected, configuration)



from betty.ancestry import Person, Event, IdentifiableEvent, Presence, Subject, Witness, Birth, Marriage
from betty.asyncio import sync
from betty.tests import TemplateTestCase


class Test(TemplateTestCase):
    template_file = 'label/event.html.j2'

    @sync
    async def test_minimal(self):
        event = Event(Birth())
        expected = 'Birth'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_identifiable(self):
        event = IdentifiableEvent('E0', Birth())
        expected = '<a href="/event/E0/index.html">Birth</a>'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_embedded_with_identifiable(self):
        event = IdentifiableEvent('E0', Birth())
        Presence(Person('P0'), Subject(), event)
        expected = 'Birth of <span class="nn" title="This person\'s name is unknown.">n.n.</span>'
        async with self._render(data={
            'event': event,
            'embedded': True,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_description(self):
        event = Event(Birth())
        event.description = 'Something happened!'
        expected = 'Birth (Something happened!)'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_witnesses(self):
        event = Event(Birth())
        Presence(Person('P0'), Witness(), event)
        expected = 'Birth'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_person_context_as_witness(self):
        event = Event(Birth())
        person = Person('P0')
        Presence(person, Witness(), event)
        expected = 'Birth (Witness)'
        async with self._render(data={
            'event': event,
            'person_context': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_person_context_as_subject(self):
        event = Event(Birth())
        person = Person('P0')
        Presence(person, Subject(), event)
        expected = 'Birth'
        async with self._render(data={
            'event': event,
            'person_context': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_person_context_and_other_as_subject(self):
        event = Event(Marriage())
        person = Person('P0')
        other_person = Person('P1')
        Presence(person, Subject(), event)
        Presence(other_person, Subject(), event)
        expected = 'Marriage with <a href="/person/P1/index.html"><span class="nn" title="This person\'s name is unknown.">n.n.</span></a>'
        async with self._render(data={
            'event': event,
            'person_context': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_subjects(self):
        event = Event(Birth())
        Presence(Person('P0'), Subject(), event)
        Presence(Person('P1'), Subject(), event)
        expected = 'Birth of <a href="/person/P0/index.html"><span class="nn" title="This person\'s name is unknown.">n.n.</span></a>, <a href="/person/P1/index.html"><span class="nn" title="This person\'s name is unknown.">n.n.</span></a>'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_without_subjects(self):
        event = Event(Birth())
        expected = 'Birth'
        async with self._render(data={
            'event': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_resource(self):
        event = Event(Birth())
        expected = 'Birth'
        async with self._render(data={
            'resource': event,
        }) as (actual, _):
            self.assertEqual(expected, actual)



import asyncio
from typing import Dict, Iterable, Optional

from betty.ancestry import Person, Place, File, Resource
from betty.site import Site


class Index:
    def __init__(self, site: Site):
        self._site = site

    async def build(self) -> Iterable[Dict]:
        return filter(None, await asyncio.gather(*[
            *[self._build_person(person) for person in self._site.ancestry.people.values()],
            *[self._build_place(place) for place in self._site.ancestry.places.values()],
            *[self._build_file(file) for file in self._site.ancestry.files.values()],
        ]))

    async def _render_resource(self, resource: Resource):
        return await self._site.jinja2_environment.get_template('search/result-%s.html.j2' % resource.resource_type_name()).render_async({
            resource.resource_type_name(): resource,
        })

    async def _build_person(self, person: Person) -> Optional[Dict]:
        if person.private:
            return
        names = []
        for name in person.names:
            if name.individual is not None:
                names.append(name.individual.lower())
            if name.affiliation is not None:
                names.append(name.affiliation.lower())
        if names:
            return {
                'text': ' '.join(names),
                'result': await self._render_resource(person),
            }

    async def _build_place(self, place: Place) -> Optional[Dict]:
        return {
            'text': ' '.join(map(lambda x: x.name.lower(), place.names)),
            'result': await self._render_resource(place),
        }

    async def _build_file(self, file: File) -> Optional[Dict]:
        if file.description is not None:
            return {
                'text': file.description.lower(),
                'result': await self._render_resource(file),
            }



from typing import Iterable, Any, Tuple, Set, Dict

Vertex = Any
Edge = Tuple[Vertex, Vertex]
Graph = Dict[Vertex, Set[Vertex]]


class GraphError(BaseException):
    pass  # pragma: no cover


class CyclicGraphError(GraphError):
    pass  # pragma: no cover


def tsort(graph: Graph) -> Iterable[Vertex]:
    edges = list(_graph_to_edges(graph))
    sorted_vertices = []
    outdegree_vertices = list(
        [edge[0] for edge in edges if not _is_target(edge[0], edges)])
    while outdegree_vertices:
        outdegree_vertex = outdegree_vertices.pop()
        if outdegree_vertex not in sorted_vertices:
            sorted_vertices.append(outdegree_vertex)
        outdegree_vertex_edges = list(
            [edge for edge in edges if edge[0] == outdegree_vertex])
        while outdegree_vertex_edges:
            edge = outdegree_vertex_edges.pop()
            edges.remove(edge)
            if not _is_target(edge[1], edges):
                outdegree_vertices.append(edge[1])
    if edges:
        raise CyclicGraphError
    isolated_vertices = list(graph.keys() - sorted_vertices)
    return sorted_vertices + isolated_vertices


def _graph_to_edges(graph: Graph) -> Iterable[Edge]:
    for from_vertex, to_vertices in graph.items():
        for to_vertex in to_vertices:
            yield from_vertex, to_vertex


def _is_target(vertex: Vertex, edges: Iterable[Edge]) -> bool:
    for edge in edges:
        if vertex == edge[1]:
            return True
    return False



import asyncio
import hashlib
import logging
import re
from contextlib import suppress
from json import load
from os.path import dirname, join, getmtime
from time import time
from typing import Optional, Dict, Callable, Tuple, Iterable, Set, Any

import aiohttp
from babel import parse_locale
from jinja2 import contextfilter
from jinja2.runtime import resolve_or_missing

from betty.ancestry import Link, HasLinks, Resource
from betty.fs import makedirs
from betty.jinja2 import Jinja2Provider
from betty.locale import Localized, negotiate_locale
from betty.media_type import MediaType
from betty.parse import PostParser
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.site import Site


class WikipediaError(BaseException):
    pass  # pragma: no cover


class NotAnEntryError(WikipediaError, ValueError):
    pass  # pragma: no cover


class RetrievalError(WikipediaError, RuntimeError):
    pass  # pragma: no cover


_URL_PATTERN = re.compile(r'^https?://([a-z]+)\.wikipedia\.org/wiki/([^/?#]+).*$')


def parse_url(url: str) -> Tuple[str, str]:
    match = _URL_PATTERN.fullmatch(url)
    if match is None:
        raise NotAnEntryError
    return match.groups()


class Entry(Localized):
    def __init__(self, locale: str, name: str, title: str, content: str):
        Localized.__init__(self, locale)
        self._name = name
        self._title = title
        self._content = content

    @property
    def name(self) -> str:
        return self._name

    @property
    def url(self) -> str:
        return 'https://%s.wikipedia.org/wiki/%s' % (self.locale, self._name)

    @property
    def title(self) -> str:
        return self._title

    @property
    def content(self) -> str:
        return self._content


class Retriever:
    def __init__(self, session: aiohttp.ClientSession, cache_directory_path: str, ttl: int = 86400):
        self._cache_directory_path = join(cache_directory_path, 'wikipedia')
        makedirs(self._cache_directory_path)
        self._ttl = ttl
        self._session = session

    async def _request(self, url: str) -> Dict:
        cache_file_path = join(self._cache_directory_path,
                               hashlib.md5(url.encode('utf-8')).hexdigest())

        response_data = None
        with suppress(FileNotFoundError):
            if getmtime(cache_file_path) + self._ttl > time():
                with open(cache_file_path) as f:
                    response_data = load(f)

        if response_data is None:
            logger = logging.getLogger()
            try:
                async with self._session.get(url) as response:
                    response_data = await response.json()
                    with open(cache_file_path, 'w') as f:
                        f.write(await response.text())
            except aiohttp.ClientError as e:
                logger.warning('Could not successfully connect to Wikipedia at %s: %s' % (url, e))
            except ValueError as e:
                logger.warning('Could not parse JSON content from Wikipedia at %s: %s' % (url, e))

        if response_data is None:
            try:
                with open(cache_file_path) as f:
                    response_data = load(f)
            except FileNotFoundError:
                raise RetrievalError('Could neither fetch %s, nor find an old version in the cache.' % url)

        return response_data

    async def _get_page_data(self, url: str) -> Dict:
        response_data = await self._request(url)
        try:
            return response_data['query']['pages'][0]
        except (LookupError, TypeError) as e:
            raise RetrievalError('Could not successfully parse the JSON format returned by %s: %s' % (url, e))

    async def get_translations(self, entry_language: str, entry_name: str) -> Dict[str, str]:
        url = 'https://%s.wikipedia.org/w/api.php?action=query&titles=%s&prop=langlinks&lllimit=500&format=json&formatversion=2' % (
            entry_language, entry_name)
        page_data = await self._get_page_data(url)
        try:
            translations_data = page_data['langlinks']
        except KeyError:
            # There may not be any translations.
            return {}
        return {translation_data['lang']: translation_data['title'] for translation_data in translations_data}

    async def get_entry(self, language: str, name: str) -> Entry:
        url = 'https://%s.wikipedia.org/w/api.php?action=query&titles=%s&prop=extracts&exintro&format=json&formatversion=2' % (
            language, name)
        page_data = await self._get_page_data(url)
        try:
            return Entry(language, name, page_data['title'], page_data['extract'])
        except KeyError as e:
            raise RetrievalError('Could not successfully parse the JSON content returned by %s: %s' % (url, e))


class _Populator:
    def __init__(self, site: Site, retriever: Retriever):
        self._site = site
        self._retriever = retriever

    async def populate(self) -> None:
        locales = set(self._site.configuration.locales)
        await asyncio.gather(*[self._populate_resource(resource, locales) for resource in self._site.ancestry.resources])

    async def _populate_resource(self, resource: Resource, locales: Set[str]) -> None:
        if not isinstance(resource, HasLinks):
            return

        entry_links = set()
        for link in resource.links:
            try:
                entry_language, entry_name = parse_url(link.url)
                entry_links.add((entry_language, entry_name))
            except NotAnEntryError:
                continue

            entry = None
            if link.label is None:
                with suppress(RetrievalError):
                    entry = await self._retriever.get_entry(entry_language, entry_name)
            await self.populate_link(link, entry_language, entry)

        for entry_language, entry_name in list(entry_links):
            entry_translations = await self._retriever.get_translations(entry_language, entry_name)
            if len(entry_translations) == 0:
                continue
            entry_languages = list(entry_translations.keys())
            for locale in locales.difference({entry_language}):
                added_entry_language = negotiate_locale(locale, entry_languages)
                if added_entry_language is None:
                    continue
                added_entry_name = entry_translations[added_entry_language]
                if (added_entry_language, added_entry_name) in entry_links:
                    continue
                try:
                    added_entry = await self._retriever.get_entry(added_entry_language, added_entry_name)
                except RetrievalError:
                    continue
                added_link = Link(added_entry.url)
                await self.populate_link(added_link, added_entry_language, added_entry)
                resource.links.add(added_link)
                entry_links.add((added_entry_language, added_entry_name))

    async def populate_link(self, link: Link, entry_language: str, entry: Optional[Entry] = None) -> None:
        if link.url.startswith('http:'):
            link.url = 'https:' + link.url[5:]
        if link.media_type is None:
            link.media_type = MediaType('text/html')
        if link.relationship is None:
            link.relationship = 'external'
        if link.locale is None:
            link.locale = entry_language
        if link.description is None:
            # There are valid reasons for links in locales that aren't supported.
            with suppress(ValueError):
                async with self._site.with_locale(link.locale):
                    link.description = _('Read more on Wikipedia.')
        if entry is not None and link.label is None:
            link.label = entry.title


class Wikipedia(Plugin, Jinja2Provider, PostParser):
    def __init__(self, site: Site):
        self._site = site

    async def __aenter__(self):
        self._session = aiohttp.ClientSession(connector=aiohttp.TCPConnector(limit_per_host=5))
        self._retriever = Retriever(self._session, join(self._site.configuration.cache_directory_path, self.name()))
        self._populator = _Populator(self._site, self._retriever)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._session.close()

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(site)

    async def post_parse(self) -> None:
        await self._populator.populate()

    @property
    def filters(self) -> Dict[str, Callable]:
        return {
            'wikipedia': self._filter_wikipedia_links,
        }

    @contextfilter
    async def _filter_wikipedia_links(self, context, links: Iterable[Link]) -> Iterable[Entry]:
        locale = parse_locale(resolve_or_missing(context, 'locale'), '-')[0]
        return filter(None, await asyncio.gather(*[self._filter_wikipedia_link(locale, link) for link in links]))

    async def _filter_wikipedia_link(self, locale: str, link: Link) -> Optional[Entry]:
        try:
            entry_language, entry_name = parse_url(link.url)
        except NotAnEntryError:
            return
        if negotiate_locale(locale, [entry_language]) is None:
            return
        try:
            return await self._retriever.get_entry(entry_language, entry_name)
        except RetrievalError:
            return

    @property
    def assets_directory_path(self) -> Optional[str]:
        return '%s/assets' % dirname(__file__)
